import conllu
import torch
import torch.nn as nn
import os
import shutil
import random
import re
import argparse


from flair.data import Dictionary, Sentence
from lib.dropout import WordDropout,LockedDropout
from transformers import BertModel,BertTokenizerFast,BertConfig
from random import sample
from collections import defaultdict
from lib.crfutils.crf import CRF
from lib.crfutils.viterbi import ViterbiDecoder,ViterbiLoss
from lib.reorder_sgml import reorder
from lib.tt2conll import conllize

from time import time

#os.environ["PYTORCH_CUDA_ALLOC_CONF"] = "max_split_size_mb:256"

class Score:
        def __init__(self, gold_total, system_total, correct, aligned_total=None):
            self.correct = correct
            self.gold_total = gold_total
            self.system_total = system_total
            self.aligned_total = aligned_total
            self.precision = correct / system_total if system_total else 0.0
            self.recall = correct / gold_total if gold_total else 0.0
            self.f1 = 2 * correct / (system_total + gold_total) if system_total + gold_total else 0.0
            self.aligned_accuracy = correct / aligned_total if aligned_total else aligned_total


class UDSpan:
    def __init__(self, start, end):
        self.start = start
        # Note that self.end marks the first position **after the end** of span,
        # so we can use characters[start:end] or range(start, end).
        self.end = end


class MTLModel(nn.Module):
    def __init__(self,sbdrnndim=256,posrnndim=512,morphrnndim=512,sbdrnnnumlayers=1,posrnnnumlayers=1,morphrnnnumlayers=1,posfflayerdim=512,morphfflayerdim=512,sbdrnnbidirectional=True,posrnnbidirectional=True,morphrnnbidirectional=True,sbdencodertype='lstm',sbdfflayerdim=256,posencodertype='lstm',morphencodertype='lstm',batchsize=32,sequencelength=256,dropout=0.0,wordropout=0.05,lockeddropout=0.5,cpu=False):
        super(MTLModel,self).__init__()

        if cpu == False:
            self.device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
        else:
            self.device = "cpu"

        # tagsets - amend labels here
        self.postagset = {'ADJ':0, 'ADP':1, 'ADV':2, 'AUX':3, 'CCONJ':4, 'DET':5, 'INTJ':6, 'NOUN':7, 'NUM':8, 'PRON':9, 'PROPN':10, 'PUNCT':11, 'SCONJ':12, 'SYM':13, 'VERB':14, 'X':15} # derived from HTB and IAHLTWiki trainsets #TODO: add other UD tags?
        self.sbd_tag2idx = {'B-SENT': 1,'O': 0}
        self.supertokenset = {'O':0,'B':1,'I':2,'E':3}

        # POS tagset in Dictionary object for Flair CRF
        self.postagsetcrf = Dictionary()
        for key in self.postagset.keys():
            self.postagsetcrf.add_item(key.strip())
        self.postagsetcrf.add_item("<START>")
        self.postagsetcrf.add_item("<STOP>")

        # FEATS dictionary
        # IMPORTANT: This should be sorted by key
        self.featstagset = {'Definite=Com':0, 'Definite=Cons':1, 'Definite=Def':2, 'Definite=Ind':3, 'Definite=Spec':4,
                            'Gender=Fem':5, 'Gender=Masc':6, 'HebBinyan=HIFIL':7, 'HebBinyan=HITPAEL':8, 'HebBinyan=HUFAL':9, 'HebBinyan=NIFAL':10,
                            'HebBinyan=NITPAEL':11, 'HebBinyan=PAAL':12, 'HebBinyan=PIEL':13, 'HebBinyan=PUAL':14, 'Number=Dual':15, 'Number=Plur':16, 'Number=Sing':17,
                            'Tense=Fut':18, 'Tense=Past':19, 'Tense=Pres':20,'VerbForm=Inf':21, 'VerbForm=Part':22,'Voice=Act':23, 'Voice=Mid':24, 'Voice=Pass':25}
        self.idxtofeatstagset = {v: k for k, v in self.featstagset.items()}
        # shared hyper-parameters
        self.sequence_length = sequencelength
        self.batch_size = batchsize

        # Embedding parameters and model
        config = BertConfig.from_pretrained('onlplab/alephbert-base',output_hidden_states=True)
        self.tokenizer = BertTokenizerFast.from_pretrained('onlplab/alephbert-base')
        self.model = BertModel.from_pretrained('onlplab/alephbert-base',config=config).to(self.device)
        self.embeddingdim = 768
        self.lastn = 4

        for param in self.model.base_model.parameters():
            param.requires_grad = False

        # Bi-LSTM Encoder for SBD
        self.sbdrnndim = sbdrnndim
        self.sbdrnnnumlayers = sbdrnnnumlayers
        self.sbdrnnbidirectional = sbdrnnbidirectional

        #Bi-LSTM Encoder for POS tagging
        self.posrnndim = posrnndim
        self.posrnnnumlayers = posrnnnumlayers
        self.posrnnbidirectional = posrnnbidirectional

        # Encoder for feats
        self.morphrnndim = morphrnndim
        self.morphrnnnumlayers = morphrnnnumlayers
        self.morphrnnbidirectional = morphrnnbidirectional

        if sbdencodertype == 'lstm':
            self.sbdencoder = nn.LSTM(input_size=self.embeddingdim, hidden_size=self.sbdrnndim // 2,
                                 num_layers=self.sbdrnnnumlayers, bidirectional=self.sbdrnnbidirectional,
                                 batch_first=True).to(self.device)
        elif sbdencodertype == 'gru':
            self.sbdencoder = nn.GRU(input_size=self.embeddingdim, hidden_size=self.sbdrnndim // 2,
                                   num_layers=self.sbdrnnnumlayers, bidirectional=self.sbdrnnbidirectional,
                                   batch_first=True).to(self.device)

        # param init
        for name, param in self.sbdencoder.named_parameters():
            try:
                if 'bias' in name:
                    nn.init.constant_(param,0.0)
                elif 'weight' in name:
                    nn.init.xavier_uniform_(param)
            except ValueError as ex:
                nn.init.constant_(param,0.0)

        if posencodertype == 'lstm':
            self.posencoder = nn.LSTM(input_size=self.embeddingdim +  len(self.supertokenset) + 1, hidden_size=self.posrnndim // 2,
                                 num_layers=self.posrnnnumlayers, bidirectional=self.posrnnbidirectional,
                                 batch_first=True).to(self.device)
        elif posencodertype == 'gru':
            self.posencoder = nn.GRU(input_size=self.embeddingdim +  len(self.supertokenset) + 1 , hidden_size=self.posrnndim // 2,
                                   num_layers=self.posrnnnumlayers, bidirectional=self.posrnnbidirectional,
                                   batch_first=True).to(self.device)

        # param init
        for name, param in self.posencoder.named_parameters():
            try:
                if 'bias' in name:
                    nn.init.constant_(param, 0.0)
                elif 'weight' in name:
                    nn.init.xavier_uniform_(param)
            except ValueError as ex:
                nn.init.constant_(param, 0.0)

        if morphencodertype == 'lstm':
            self.morphencoder = nn.LSTM(input_size=self.embeddingdim + len(self.postagsetcrf) + len(self.supertokenset) + 1, hidden_size=self.morphrnndim // 2,
                                 num_layers=self.morphrnnnumlayers, bidirectional=self.morphrnnbidirectional,
                                 batch_first=True).to(self.device)
        elif morphencodertype == 'gru':
            self.morphencoder = nn.GRU(input_size=self.embeddingdim + len(self.postagsetcrf) + len(self.supertokenset) + 1, hidden_size=self.morphrnndim // 2,
                                   num_layers=self.morphrnnnumlayers, bidirectional=self.morphrnnbidirectional,
                                   batch_first=True).to(self.device)

        # param init
        for name, param in self.morphencoder.named_parameters():
            try:
                if 'bias' in name:
                    nn.init.constant_(param, 0.0)
                elif 'weight' in name:
                    nn.init.xavier_uniform_(param)
            except ValueError as ex:
                nn.init.constant_(param, 0.0)


        self.relu = nn.ReLU()

        # Reproject embeddings layer
        self.sbdembedding2nn = nn.Linear(in_features=self.embeddingdim ,out_features=self.embeddingdim).to(self.device)
        self.sbdfflayerdim = sbdfflayerdim
        self.sbdfflayer = nn.Linear(in_features=self.sbdrnndim, out_features=self.sbdfflayerdim).to(self.device)

        # param init
        for name, param in self.sbdembedding2nn.named_parameters():
            if 'bias' in name:
                nn.init.constant_(param, 0.0)
            elif 'weight' in name:
                nn.init.xavier_normal_(param)

        for name, param in self.sbdfflayer.named_parameters():
            if 'bias' in name:
                nn.init.constant_(param, 0.0)
            elif 'weight' in name:
                nn.init.xavier_normal_(param)


        # Intermediate feedforward layer
        self.posembedding2nn = nn.Linear(in_features=self.embeddingdim + len(self.supertokenset) + 1,out_features=self.embeddingdim  +  len(self.supertokenset) + 1 ).to(self.device)
        self.posfflayerdim = posfflayerdim
        self.posfflayer = nn.Linear(in_features=self.posrnndim, out_features=self.posfflayerdim).to(self.device)

        # param init
        for name, param in self.posembedding2nn.named_parameters():
            if 'bias' in name:
                nn.init.constant_(param, 0.0)
            elif 'weight' in name:
                nn.init.xavier_normal_(param)

        for name, param in self.posfflayer.named_parameters():
            if 'bias' in name:
                nn.init.constant_(param, 0.0)
            elif 'weight' in name:
                nn.init.xavier_normal_(param)

        # Intermediate feedforward layer
        self.morphembedding2nn = nn.Linear(in_features=self.embeddingdim + len(self.postagsetcrf) + len(self.supertokenset) + 1 ,
                                         out_features=self.embeddingdim + len(self.postagsetcrf) + len(self.supertokenset) + 1).to(self.device)
        self.morphfflayerdim = morphfflayerdim
        self.morphfflayer = nn.Linear(in_features=self.morphrnndim, out_features=self.morphfflayerdim).to(self.device)

        # param init
        for name, param in self.morphembedding2nn.named_parameters():
            if 'bias' in name:
                nn.init.constant_(param, 0.0)
            elif 'weight' in name:
                nn.init.xavier_normal_(param)

        # param init
        for name, param in self.morphfflayer.named_parameters():
            if 'bias' in name:
                nn.init.constant_(param, 0.0)
            elif 'weight' in name:
                nn.init.xavier_normal_(param)

        # Label space for the pos tagger
        self.hidden2postag = nn.Linear(in_features=self.posfflayerdim,out_features=len(self.postagsetcrf)).to(self.device)
        for name, param in self.hidden2postag.named_parameters():
            if 'bias' in name:
                nn.init.constant_(param, 0.0)
            elif 'weight' in name:
                nn.init.xavier_normal_(param)

        # Label space for sent splitter
        self.hidden2sbd = nn.Linear(in_features=self.sbdfflayerdim,out_features=len(self.sbd_tag2idx.keys())).to(self.device)
        # param init
        for name, param in self.hidden2sbd.named_parameters():
            if 'bias' in name:
                nn.init.constant_(param, 0.0)
            elif 'weight' in name:
                nn.init.xavier_normal_(param)


        # label space for morph feats
        self.hidden2feats = nn.Linear(in_features=self.morphfflayerdim,out_features=len(self.featstagset)).to(self.device)
        for name, param in self.hidden2feats.named_parameters():
            if 'bias' in name:
                nn.init.constant_(param, 0.0)
            elif 'weight' in name:
                nn.init.xavier_normal_(param)

        self.dropout = nn.Dropout(dropout)
        self.worddropout = WordDropout(wordropout)
        self.lockeddropout = LockedDropout(lockeddropout)

        self.poscrf = CRF(self.postagsetcrf,len(self.postagsetcrf),init_from_state_dict=False,cpu=cpu).to(self.device) # TODO: parameterize
        self.viterbidecoder = ViterbiDecoder(self.postagsetcrf,cpu=cpu)

        for name, param in self.poscrf.named_parameters():
            if 'bias' in name:
                nn.init.constant_(param, 0.0)
            elif 'weight' in name:
                nn.init.xavier_normal_(param)

        self.stride_size = 10
        self.sigmoid = nn.Sigmoid()

    def shingle(self,toks,labels=None):
        """
        Returns the span embeddings, labelspans, and 'final mapping'
        """
        spans = []
        labelspans = []
        final_mapping = {}

        # Hack tokens up into overlapping shingles
        wraparound = torch.cat((toks[-self.stride_size:],toks,toks[: self.sequence_length]),dim=0)
        if labels:
            labelwraparound = labels[-self.stride_size:] + labels + labels[: self.sequence_length]
        idx = 0
        mapping = defaultdict(set)
        snum = 0
        while idx < len(toks):
            if idx + self.sequence_length < len(wraparound):
                span = wraparound[idx: idx + self.sequence_length]
                if labels:
                    labelspan = labelwraparound[idx: idx + self.sequence_length]
            else:
                span = wraparound[idx:]
                if labels:
                    labelspan = labelwraparound[idx:]

            spans.append(span)
            if labels:
                labelspans.append(labelspan)

            for i in range(idx - self.stride_size, idx + self.sequence_length - self.stride_size):
                # start, end, snum
                if i >= 0 and i < len(toks):
                    mapping[i].add(
                        (idx - self.stride_size, idx + self.sequence_length - self.stride_size, snum))
            idx += self.stride_size
            snum += 1

        for idx in mapping:
            best = self.sequence_length
            for m in mapping[idx]:
                start, end, snum = m
                dist_to_end = end - idx
                dist_to_start = idx - start
                delta = abs(dist_to_end - dist_to_start)
                if delta < best:
                    best = delta
                    final_mapping[idx] = (snum, idx - start)  # Get sentence number and position in sentence

        spans = torch.stack(spans)

        return spans,labelspans,final_mapping

    def forward(self,data,mode='train'):

        badrecords = [] # stores records where AlephBERT's tokenization 'messed up' the sentence's sequence length, and removes these sentences from the batch.
        featslabels = None

        # Extract the sentences and labels
        if mode == 'train': # training is on a batch, so 3D tensor
            sentences = [' '.join([s.split('\t')[0].strip() for s in sls]) for sls in data]
            sbdlabels = [[self.sbd_tag2idx[s.split('\t')[2].strip()] for s in sls] for sls in data]
            poslabels = [[self.postagsetcrf.get_idx_for_item(s.split('\t')[1].strip()) for s in sls] for sls in data]

            supertokenlabels = []
            featslabels = []

            for sls in data:
                record = []
                featsrecord = []
                for s in sls:
                    temp = [0] * len(self.supertokenset)
                    temp[self.supertokenset[s.split('\t')[3].strip()]] = 1
                    record.append(temp)

                    tempfeats = [0] * len(self.featstagset)
                    fts = s.split('\t')[-1].strip()
                    if fts != '':
                        fts = fts.split('|')
                        for f in fts:
                            key = f.split('=')[0]
                            value = f.split('=')[1]
                            if ',' not in value:
                                if f not in self.featstagset.keys(): continue
                                tempfeats[self.featstagset[f]] = 1
                            else:
                                value = value.split(',')
                                for v in value:
                                    if key + '=' + v not in self.featstagset.keys(): continue
                                    tempfeats[self.featstagset[key + '=' + v]] = 1

                    featsrecord.append(tempfeats)

                supertokenlabels.append(record)
                featslabels.append(featsrecord)

        elif mode == 'dev': # inference is on a single record
            sentences = [' '.join([s.split('\t')[0].strip() for s in data])]
            sbdlabels = [self.sbd_tag2idx[s.split('\t')[2].strip()] for s in data]
            poslabels = [self.postagsetcrf.get_idx_for_item(s.split('\t')[1].strip()) for s in data]

            supertokenlabels = []
            featslabels = []
            for s in data:
                temp = [0] * len(self.supertokenset)
                temp[self.supertokenset[s.split('\t')[3].strip()]] = 1
                supertokenlabels.append(temp)

                tempfeats = [0] * len(self.featstagset)
                fts = s.split('\t')[-1].strip()
                if fts != '':
                    fts = fts.split('|')
                    for f in fts:
                        key = f.split('=')[0]
                        value = f.split('=')[1]
                        if ',' not in value:
                            if f not in self.featstagset.keys(): continue
                            tempfeats[self.featstagset[f]] = 1
                        else:
                            value = value.split(',')
                            for v in value:
                                if key + '=' + v not in self.featstagset.keys(): continue
                                tempfeats[self.featstagset[key + '=' + v]] = 1

                featslabels.append(tempfeats)

        else: # test - a tuple of text and supertoken labels
            sentences = [' '.join([s.split('\t')[0].strip() for s in data[0]])]

            supertokenlabels = []
            for s in data[1]:
                temp = [0] * len(self.supertokenset)
                temp[self.supertokenset[s.strip()]] = 1
                supertokenlabels.append(temp)

            sbdlabels = None
            poslabels = None

        # Make embeddings and scalar average them across subwords, vertically.
        sentences = [d.split() for d in sentences] # for AlephBERT
        tokens = self.tokenizer(sentences,return_tensors='pt',padding=True,is_split_into_words=True,truncation=True).to(self.device) # tell AlephBERT that there is some tokenization already.

        output = self.model(**tokens)
        hiddenstates = output[2][-self.lastn:]
        scalarsum = hiddenstates[0]
        for i in range(1,self.lastn):
            scalarsum = torch.add(scalarsum,hiddenstates[i],alpha=1)

        embeddings = torch.div(scalarsum,self.lastn)
        #embeddings = embeddings.to(self.device)

        """
        Average the subword embeddings within the horizontal sequence.
        This process will drop the [CLS],[SEP] and [PAD] tokens
        """

        avgembeddings = []
        for k in range(0,len(tokens.encodings)):
            emb = []
            maxindex = max([w for w in tokens.encodings[k].words if w is not None])

            try:
                assert maxindex == self.sequence_length - 1  # otherwise won't average correctly and align with labels
            except AssertionError:
                #print ('max index not equal sequence len. Default labels will be applied.')
                badrecords.append(k)
                continue

            for i in range(0,self.sequence_length):

                indices = [j for j,x in enumerate(tokens.encodings[k].words) if x == i]
                if len(indices) == 0: # This strange case needs to be handled.
                    emb.append(torch.zeros(self.embeddingdim ,device=self.device))
                elif len(indices) == 1: # no need to average
                    emb.append(embeddings[k][indices[0]])
                else: # needs to aggregate - average
                    slice = embeddings[k][indices[0]:indices[-1] + 1]
                    slice = torch.mean(input=slice,dim=0,keepdim=False)
                    emb.append(slice)
            try:
                assert len(emb) == self.sequence_length # averaging was correct and aligns with the labels
            except AssertionError:
                #print ('embedding not built correctly. Default labels will be applied')
                badrecords.append(k)
                continue

            emb = torch.stack(emb)
            avgembeddings.append(emb)

        badrecords = sorted(badrecords,reverse=True)

        avgembeddings = torch.stack(avgembeddings)
        for record in badrecords:
            sbdlabels.pop(record)
            poslabels.pop(record)
            self.batch_size -= 1

        if mode != 'train':
            if avgembeddings.size(dim=1) > self.stride_size: # don't shingle if seqlen less than the overlap
                # squeeze the embedding, as it's a single sentence
                avgembeddings = torch.squeeze(avgembeddings)
                # shingle the sentence embedding and its label, to calculate the dev loss later
                sbdembeddings,finalsbdlabels,finalmapping = self.shingle(avgembeddings,sbdlabels)
                # restore dimensionality for the POS tagging pipeline.
                avgembeddings = torch.unsqueeze(avgembeddings,dim=0)
            else:
                sbdembeddings = avgembeddings
                finalsbdlabels = sbdlabels
                finalmapping = None
        else:
            sbdembeddings = avgembeddings
            finalsbdlabels = sbdlabels
            finalmapping = None

        sbdembeddings = self.dropout(sbdembeddings)
        sbdembeddings = self.worddropout(sbdembeddings)
        sbdembeddings = self.lockeddropout(sbdembeddings)

        sbdembeddings = self.sbdembedding2nn(sbdembeddings)

        # SBD encoder and labels
        feats, _ = self.sbdencoder(sbdembeddings)
        feats = self.sbdfflayer(feats)
        feats = self.relu(feats)
        feats = self.dropout(feats)
        feats = self.lockeddropout(feats)

        # SBD logits
        sbdlogits = self.hidden2sbd(feats)

        #get the sbd predictions as input to the POS encoder
        if mode == 'train':
            sbdpreds = torch.argmax(sbdlogits,dim=2,keepdim=True)
        else:
            # Predict from the shingles for SBD.
            # 'Believe the span where the token is most in the middle'
            if sbdlogits.size(dim=1) > self.stride_size:
                sbdpreds = []
                for idx in finalmapping:
                    snum, position = finalmapping[idx]
                    label = torch.argmax(sbdlogits[snum][position]).item()
                    sbdpreds.append(label)

                # Unsqueeze for input to the POS Encoder
                sbdpreds = torch.LongTensor(sbdpreds)
                sbdpreds = torch.unsqueeze(sbdpreds, dim=0)
                sbdpreds = torch.unsqueeze(sbdpreds, dim=2)
                sbdpreds = sbdpreds.to(self.device)
            else:
                sbdpreds = torch.argmax(sbdlogits, dim=2, keepdim=True)

        supertokenlabels = torch.LongTensor(supertokenlabels)
        supertokenlabels = supertokenlabels.to(self.device)
        if mode in ('dev','test'):
            supertokenlabels = torch.unsqueeze(supertokenlabels,dim=0)

        # Add the SBD predictions to the POS Encoder Input!
        posembeddings = torch.cat((avgembeddings,sbdpreds,supertokenlabels),dim=2)

        posembeddings = self.dropout(posembeddings)
        posembeddings = self.worddropout(posembeddings)
        posembeddings = self.lockeddropout(posembeddings)
        posembeddings = self.posembedding2nn(posembeddings)

        feats,_ = self.posencoder(posembeddings)
        feats = self.posfflayer(feats)
        feats = self.relu(feats)
        feats = self.dropout(feats)
        feats = self.lockeddropout(feats)

        # logits for pos
        poslogits = self.hidden2postag(feats)
        poslogits = self.poscrf(poslogits)

        # get the pos CRF predictions
        if mode == 'train':
            lengths = [self.sequence_length] * self.batch_size
        else:
            lengths = [self.sequence_length]
        scores = (poslogits, lengths, self.poscrf.transitions)
        sents = []
        for s in sentences:
            sents.append(Sentence(' '.join(s),use_tokenizer=False))

        pospreds = self.viterbidecoder.decode(scores, False, sents)
        pospreds = [[self.postagsetcrf.get_idx_for_item(p[0])for p in pr] for pr in pospreds[0]]
        pospredsonehot = []
        for pred in pospreds:
            preds = []
            for p in pred:
                onehot = [0] * len(self.postagsetcrf)
                onehot[p] = 1
                preds.append(onehot)
            pospredsonehot.append(preds)

        pospredsonehot = torch.LongTensor(pospredsonehot)
        pospredsonehot = pospredsonehot.to(self.device)

        morphembeddings = torch.cat((avgembeddings, sbdpreds, supertokenlabels,pospredsonehot), dim=2)
        morphembeddings = self.dropout(morphembeddings)
        morphembeddings = self.worddropout(morphembeddings)
        morphembeddings = self.lockeddropout(morphembeddings)
        morphembeddings = self.morphembedding2nn(morphembeddings)

        feats, _ = self.morphencoder(morphembeddings)
        feats = self.morphfflayer(feats)
        feats = self.relu(feats)
        feats = self.dropout(feats)
        feats = self.lockeddropout(feats)

        # logits for morphs
        featslogits = self.hidden2feats(feats)

        if mode in ('dev','test'):
            # Squeeze these to return to the Trainer for scores, now that we are done with them
            sbdpreds = torch.squeeze(sbdpreds,dim=2)
            sbdpreds = torch.squeeze(sbdpreds, dim=0)
            sbdpreds = sbdpreds.tolist()

            # Unroll the pos predictions
            pospreds = [p for pred in pospreds for p in pred]

        else:
            sbdpreds = None
            pospreds = None

        return sbdlogits, finalsbdlabels, sbdpreds, poslogits, poslabels, pospreds, featslogits,featslabels # returns the logits and labels

class Tagger():
    def __init__(self,trainflag=False,trainfile=None,devfile=None,testfile=None,sbdrnndim=256,sbdrnnnumlayers=1,sbdrnnbidirectional=True,sbdfflayerdim=256,posrnndim=512,posrnnnumlayers=1,posrnnbidirectional=True,posfflayerdim=512,morphrnndim=512,morphrnnnumlayers=1,morphrnnbidirectional=True,morphfflayerdim=512,morphencodertype='lstm',dropout=0.05,wordropout=0.05,lockeddropout=0.5,sbdencodertype='lstm',posencodertype='lstm',learningrate = 0.001,bestmodelpath='../data/checkpoint/',batchsize=32,sequencelength=256,datatype='htb',cpu=False):

        self.mtlmodel = MTLModel(sbdrnndim=sbdrnndim,sbdrnnnumlayers=sbdrnnnumlayers,sbdrnnbidirectional=sbdrnnbidirectional,sbdencodertype=sbdencodertype,sbdfflayerdim=sbdfflayerdim,dropout=dropout,wordropout=wordropout,lockeddropout=lockeddropout,posrnndim=posrnndim,posrnnbidirectional=posrnnbidirectional,posencodertype=posencodertype,posrnnnumlayers=posrnnnumlayers,posfflayerdim=posfflayerdim,morphrnndim=morphrnndim,morphrnnnumlayers=morphrnnnumlayers,morphencodertype=morphencodertype,morphrnnbidirectional=morphrnnbidirectional,morphfflayerdim=morphfflayerdim,batchsize=batchsize,sequencelength=sequencelength,cpu=cpu)

        if trainflag == True:

            from torch.utils.tensorboard import SummaryWriter
            if os.path.isdir('../data/tensorboarddir/'):
                shutil.rmtree('../data/tensorboarddir/')
            os.mkdir('../data/tensorboarddir/')

            if not os.path.isdir(bestmodelpath):
                os.mkdir(bestmodelpath)

            self.writer = SummaryWriter('../data/tensorboarddir/')

            self.trainingdatafile = '../data/sentsplit_postag_train_gold.tab'
            self.devdatafile = '../data/sentsplit_postag_dev_gold.tab'

            self.bestmodel = bestmodelpath + datatype + '_best_mtlmodel.pt'

        if cpu == False:
            self.device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
        else:
            self.device = "cpu"

        self.trainflag = trainflag
        self.trainfile = trainfile
        self.devfile = devfile
        self.testfile = testfile

        self.learningrate = learningrate

        # Loss for pos tagging
        self.postagloss = ViterbiLoss(self.mtlmodel.postagsetcrf,cpu=cpu)
        self.postagloss.to(self.device)

        # Loss for sentence splitting
        self.sbdloss = nn.CrossEntropyLoss(weight=torch.FloatTensor([1,3]))
        self.sbdloss.to(self.device)

        self.featsloss = nn.BCEWithLogitsLoss()
        self.featsloss.to(self.device)

        self.optimizer = torch.optim.AdamW(list(self.mtlmodel.sbdencoder.parameters()) +  list(self.mtlmodel.sbdembedding2nn.parameters()) +
                                           list(self.mtlmodel.hidden2sbd.parameters()) + list(self.mtlmodel.posencoder.parameters()) + list(self.mtlmodel.posembedding2nn.parameters())
                                           + list(self.mtlmodel.hidden2postag.parameters()) + list(self.mtlmodel.poscrf.parameters())
                                           + list(self.mtlmodel.hidden2feats.parameters()) + list(self.mtlmodel.morphfflayer.parameters()) + list(self.mtlmodel.morphembedding2nn.parameters()) + list(self.mtlmodel.morphencoder.parameters())
                                           + list(self.mtlmodel.posfflayer.parameters()) + list(self.mtlmodel.sbdfflayer.parameters()), lr=learningrate)

        self.scheduler = torch.optim.lr_scheduler.CyclicLR(self.optimizer,base_lr=learningrate/10,max_lr=learningrate,step_size_up=250,cycle_momentum=False)
        self.evalstep = 50

        self.sigmoidthreshold = 0.5

    def set_seed(self, seed):

        random.seed(seed)
        torch.manual_seed(seed)

    def train(self,checkpointfile=None):

        def read_file(mode='train'):

            dataset = []
            if mode == 'dev':
                with open(self.devdatafile, 'r') as fi:
                    lines = fi.readlines()
                    #lines = list(reversed(lines))  # hebrew is right to left...

                    for idx in range(0, len(lines), self.mtlmodel.sequence_length):
                        if idx + self.mtlmodel.sequence_length >= len(lines):
                            slice = lines[idx:len(lines)]
                        else:
                            slice = lines[idx: idx + self.mtlmodel.sequence_length]

                        dataset.append(slice)

                test = [d for slice in dataset for d in slice]
                assert len(test) == len(lines)

            else:
                with open(self.trainingdatafile,'r') as fi:
                    lines = fi.readlines()
                    #lines = list(reversed(lines)) # hebrew is right to left...

                    # shingle it here to get more training data
                    for idx in range(0,len(lines),self.mtlmodel.sequence_length - self.mtlmodel.stride_size):
                        if idx + self.mtlmodel.sequence_length >= len(lines):
                            slice = lines[idx:len(lines)]
                            dataset.append(slice)
                            break
                        else:
                            slice = lines[idx: idx + self.mtlmodel.sequence_length]
                            dataset.append(slice)

            return dataset

        epochs = 5000
        bestf1 = float('-inf')

        trainingdata = read_file()
        devdata = read_file(mode='dev')

        if checkpointfile is not None:
            checkpoint = torch.load(checkpointfile)
            self.mtlmodel.load_state_dict(checkpoint['model_state_dict'])
            self.optimizer.load_state_dict(checkpoint['optimizer_state_dict'])
            self.mtlmodel.poscrf.load_state_dict(checkpoint['poscrf_state_dict'])

        self.set_seed(42)

        for epoch in range(1,epochs):

            old_batchsize = self.mtlmodel.batch_size

            self.mtlmodel.train()
            self.optimizer.zero_grad()

            data = sample(trainingdata,self.mtlmodel.batch_size)
            data = [datum for datum in data if len(datum) == self.mtlmodel.sequence_length]
            self.mtlmodel.batch_size = len(data)

            sbdlogits, sbdlabels, _, poslogits, poslabels, _ , featslogits, featslabels = self.mtlmodel(data)

            sbdtags = torch.LongTensor(sbdlabels).to(self.device)
            sbdlogits = sbdlogits.permute(0,2,1)
            sbdloss = self.sbdloss(sbdlogits,sbdtags)

            lengths = [self.mtlmodel.sequence_length] * self.mtlmodel.batch_size
            lengths = torch.LongTensor(lengths).to(self.device)
            scores = (poslogits, lengths, self.mtlmodel.poscrf.transitions)

            # unwrap the pos tags into one long list first
            postags = [p for pos in poslabels for p in pos]
            postags = torch.LongTensor(postags).to(self.device)
            posloss = self.postagloss(scores,postags)

            featstags = torch.FloatTensor(featslabels).to(self.device)
            featsloss = self.featsloss(featslogits,featstags)

            mtlloss = posloss + sbdloss + featsloss # TODO: learnable weights?
            mtlloss.backward()
            self.optimizer.step()
            self.scheduler.step()

            if old_batchsize != self.mtlmodel.batch_size:
                self.mtlmodel.batch_size = old_batchsize

            self.writer.add_scalar('train_pos_loss', posloss.item(), epoch)
            self.writer.add_scalar('train_sbd_loss', sbdloss.item(), epoch)
            self.writer.add_scalar('train_feats_loss', featsloss.item(), epoch)
            self.writer.add_scalar('train_joint_loss', mtlloss.item(), epoch)

            """""""""""""""""""""""""""""""""""""""""""""
            Do dev evaluation after evalstep number of epochs
            """""""""""""""""""""""""""""""""""""""""""""
            if epoch % self.evalstep == 0:

                self.mtlmodel.eval()

                with torch.no_grad():

                    totalsbddevloss = 0
                    totalposdevloss = 0
                    totalfeatsdevloss = 0

                    allsbdpreds = []
                    allsbdgold = []
                    allpospreds = []
                    allposgold = []
                    allfeatsgold = []
                    allfeatspreds = []

                    # because of shingling for SBD, the dev data needs to be split in slices for inference, as GPU may run out of memory with shingles on the full token list.
                    # shingling and SBD prediction is done on the individual slice, as well as POS tag predictions and feats.
                    # TODO This naturally increases prediction time...but can't think of a better way.
                    for slice in devdata:

                        old_seqlen = self.mtlmodel.sequence_length
                        if len(slice) != self.mtlmodel.sequence_length: # this will happen in one case, for the last slice in the dev batch
                            self.mtlmodel.sequence_length = len(slice)

                        # Extract gold labels
                        goldsbdlabels = [s.split('\t')[2].strip() for s in slice]
                        goldsbdlabels = [self.mtlmodel.sbd_tag2idx[s] for s in goldsbdlabels]
                        goldposlabels = [s.split('\t')[1].strip() for s in slice]
                        goldposlabels = [self.mtlmodel.postagsetcrf.get_idx_for_item(s) for s in goldposlabels]
                        goldfeatslabels = [s.split('\t')[4].strip() for s in slice]

                        # RUn through the model and get the labels and logits (and preds for stepwise models)
                        sbdlogits, sbdlabels, sbdpreds, poslogits, poslabels, pospreds, featslogits, featslabels = self.mtlmodel(slice,mode='dev')

                        # get the feats predictions
                        featspreds = self.mtlmodel.sigmoid(featslogits)
                        featspreds = (featspreds > self.sigmoidthreshold).long()
                        featspreds = torch.squeeze(featspreds).tolist()

                        # get the sbd loss
                        sbdlogits = sbdlogits.permute(0,2,1)
                        sbdtags = torch.LongTensor(sbdlabels).to(self.device)
                        sbddevloss = self.sbdloss(sbdlogits, sbdtags).item()

                        # get the pos loss
                        postags = torch.LongTensor(poslabels)
                        postags = postags.to(self.device)
                        lengths = [self.mtlmodel.sequence_length]
                        lengths = torch.LongTensor(lengths).to(self.device)
                        scores = (poslogits, lengths, self.mtlmodel.poscrf.transitions)
                        posdevloss = self.postagloss(scores,postags).item()

                        # get the feats loss
                        featstags = torch.FloatTensor(featslabels)
                        featstags = featstags.to(self.device)
                        featstags = torch.unsqueeze(featstags,dim=0)
                        featsdevloss = self.featsloss(featslogits,featstags).item()

                        # add up the losses across the slices for the avg
                        totalsbddevloss += sbddevloss
                        totalposdevloss += posdevloss
                        totalfeatsdevloss += featsdevloss

                        # build the feats tags for the sequence
                        featsslicepreds = []
                        for preds in featspreds:
                            featsstr = ''
                            for i in range(0,len(preds)):
                                if preds[i] != 0:
                                    if self.mtlmodel.idxtofeatstagset[i].split('=')[0] == featsstr.split('|')[-1].split('=')[0]:
                                        featsstr += ',' + self.mtlmodel.idxtofeatstagset[i].split('=')[1]
                                    else:
                                        if featsstr != '':
                                            featsstr = featsstr + '|' + self.mtlmodel.idxtofeatstagset[i]
                                        else:
                                            featsstr += self.mtlmodel.idxtofeatstagset[i]

                            featsslicepreds.append(featsstr)

                        # build the gold and predictions for the entire dev set
                        allsbdpreds.extend(sbdpreds)
                        allsbdgold.extend(goldsbdlabels)
                        allpospreds.extend(pospreds)
                        allposgold.extend(goldposlabels)
                        allfeatsgold.extend(goldfeatslabels)
                        allfeatspreds.extend(featsslicepreds)

                    #print ('inference time')
                    #print (time() - start)
                    if self.mtlmodel.sequence_length != old_seqlen:
                        self.mtlmodel.sequence_length = old_seqlen

                    # Now get the scores
                    goldspans = []
                    predspans = []
                    goldstartindex = 0
                    predstartindex = 0

                    for i in range(0,len(allsbdgold)):
                        if allsbdgold[i] == 1: #B-SENT
                            goldspans.append(UDSpan(goldstartindex,i))
                            goldstartindex = i
                        if allsbdpreds[i] == 1:
                            predspans.append(UDSpan(predstartindex,i))
                            predstartindex = i

                    sbdscores = self.spans_score(goldspans,predspans)

                    correctpos = sum([1 if p == g else 0 for p,g in zip(allpospreds,allposgold)])
                    posscores = Score(len(allposgold),len(allpospreds),correctpos,len(allpospreds))

                    correctfeats = sum([1 if p == g else 0 for p,g in zip(allfeatspreds,allfeatsgold)])
                    featsscores = Score(len(allfeatsgold),len(allfeatspreds),correctfeats,len(allfeatspreds))

                    # Write the scores and losses to tensorboard and console
                    mtlloss = (totalsbddevloss + totalposdevloss + totalfeatsdevloss) / len(devdata)

                    self.writer.add_scalar("mtl_dev_loss", round(mtlloss, 4),
                                           int(epoch / self.evalstep))
                    print('mtl dev loss:' + str(round((totalsbddevloss / len(devdata) + (totalposdevloss / len(devdata))), 4)))

                    self.writer.add_scalar("sbd_dev_loss",round(totalsbddevloss/len(devdata),4),int(epoch / self.evalstep))
                    self.writer.add_scalar("sbd_dev_f1", round(sbdscores.f1,4), int(epoch / self.evalstep))
                    self.writer.add_scalar("sbd_dev_precision", round(sbdscores.precision, 4), int(epoch / self.evalstep))
                    self.writer.add_scalar("sbd_dev_recall", round(sbdscores.recall, 4), int(epoch / self.evalstep))

                    print ('\n')
                    self.writer.add_scalar("pos_dev_loss", round(totalposdevloss / len(devdata), 4),
                                           int(epoch / self.evalstep))
                    self.writer.add_scalar("pos_dev_f1", round(posscores.f1, 4), int(epoch / self.evalstep))
                    self.writer.add_scalar("pos_dev_precision", round(posscores.precision, 4),
                                           int(epoch / self.evalstep))
                    self.writer.add_scalar("pos_dev_recall", round(posscores.recall, 4), int(epoch / self.evalstep))

                    self.writer.add_scalar("feats_dev_loss",round(totalfeatsdevloss / len(devdata),4), int(epoch / self.evalstep))
                    self.writer.add_scalar("feats_dev_f1",round(featsscores.f1,4),int(epoch / self.evalstep))
                    self.writer.add_scalar("feats_dev_precision",round(featsscores.precision,4),int(epoch / self.evalstep))
                    self.writer.add_scalar("feats_dev_recall", round(featsscores.recall, 4),int(epoch / self.evalstep))

                    print ('sbd dev f1:' + str(sbdscores.f1))
                    print('sbd dev precision:' + str(sbdscores.precision))
                    print('sbd dev recall:' + str(sbdscores.recall))
                    print ('\n')

                    print('pos dev f1:' + str(posscores.f1))
                    print('pos dev precision:' + str(posscores.precision))
                    print('pos dev recall:' + str(posscores.recall))
                    print('\n')

                    print('feats dev f1:' + str(featsscores.f1))
                    print('feats dev precision:' + str(featsscores.precision))
                    print('feats dev recall:' + str(featsscores.recall))
                    print('\n')

                    # save the best model
                    if (sbdscores.f1 + posscores.f1 + featsscores.f1) / 3 > bestf1:
                        bestf1 = (sbdscores.f1 + posscores.f1 + featsscores.f1) / 3
                        bestmodel = self.bestmodel.replace('.pt','_' + str(round(mtlloss,6)) + '_' + str(round(sbdscores.f1,6)) + '_' + str(round(posscores.f1,6)) + '_' + str(round(featsscores.f1,6)) + '.pt')
                        torch.save({'epoch':epoch,'model_state_dict':self.mtlmodel.state_dict(),'optimizer_state_dict':self.optimizer.state_dict(),'poscrf_state_dict':self.mtlmodel.poscrf.state_dict()},bestmodel)

    def inference(self,toks,sent_tag='auto',checkpointfile=None):

        def is_tok(sgml_line):
            return len(sgml_line) > 0 and not (sgml_line.startswith("<") and sgml_line.endswith(">"))

        def unescape(token):
            token = token.replace("&quot;", '"')
            token = token.replace("&lt;", "<")
            token = token.replace("&gt;", ">")
            token = token.replace("&amp;", "&")
            token = token.replace("&apos;", "'")
            return token

        slices = []
        toks = unescape(toks)  # Splitter is trained on UTF-8 forms, since LM embeddings know characters like '&'
        lines = toks.strip().split("\n")

        taggedsbdpreds = []

        # add super token tags
        supertokenlabels = []
        for i in range(0,len(lines)):
            if i > 0:
                prevtoken = lines[i-1]
            if i < len(lines) - 1:
                nexttoken = lines[i + 1]

            currtoken = lines[i]

            if is_tok(currtoken):
                if not is_tok(prevtoken):
                    if not is_tok(nexttoken):
                        supertokenlabels.append("O")
                    else:
                        supertokenlabels.append("B")
                else:
                    if not is_tok(nexttoken):
                        supertokenlabels.append("E")
                    else:
                        supertokenlabels.append("I")

                if sent_tag != 'auto':
                    if i != 2 and lines[i - 2] == "<" + sent_tag + ">":
                        taggedsbdpreds.append(1)
                    else:
                        taggedsbdpreds.append(0)

        toks = [l for l in lines if is_tok(l)]
        toks = [re.sub(r"\t.*", "", t) for t in toks]

        assert len(toks) == len(supertokenlabels)
        if sent_tag != 'auto':
            assert len(taggedsbdpreds) == len(toks)

        # slice up the token list into slices of seqlen for GPU RAM reasons
        for idx in range(0, len(toks), self.mtlmodel.sequence_length):
            if idx + self.mtlmodel.sequence_length >= len(toks):
                slice = toks[idx:len(toks)]
                supertokenslice = supertokenlabels[idx:len(toks)]
            else:
                slice = toks[idx: idx + self.mtlmodel.sequence_length]
                supertokenslice = supertokenlabels[idx: idx + self.mtlmodel.sequence_length]

            slices.append((slice,supertokenslice))

        test = [d for s in slices for d in s[0]]

        assert len(test) == len(toks)

        if checkpointfile is not None:

            checkpoint = torch.load(checkpointfile,map_location=self.device)
            self.mtlmodel.load_state_dict(checkpoint['model_state_dict'])
            self.optimizer.load_state_dict(checkpoint['optimizer_state_dict'])
            self.mtlmodel.poscrf.load_state_dict(checkpoint['poscrf_state_dict'])

        self.mtlmodel.eval()

        with torch.no_grad():

            allwords = []
            allsbdpreds = []
            allpospreds = []
            allfeatspreds = []

            for slice in slices:

                if len(slice[0]) != self.mtlmodel.sequence_length:  # this will happen in one case, for the last slice in the batch
                    self.mtlmodel.sequence_length = len(slice[0])

                _, _, sbdpreds, _,_,pospreds, featslogits,_ = self.mtlmodel(slice, mode='test')

                # get the feats predictions
                featspreds = self.mtlmodel.sigmoid(featslogits)
                featspreds = (featspreds > self.sigmoidthreshold).long()
                featspreds = torch.squeeze(featspreds).tolist()

                featsslicepreds = []
                for preds in featspreds:
                    featsstr = ''
                    for i in range(0, len(preds)):
                        if preds[i] != 0:
                            if self.mtlmodel.idxtofeatstagset[i].split('=')[0] == featsstr.split('|')[-1].split('=')[0]:
                                featsstr += ',' + self.mtlmodel.idxtofeatstagset[i].split('=')[1]
                            else:
                                if featsstr != '':
                                    featsstr = featsstr + '|' + self.mtlmodel.idxtofeatstagset[i]
                                else:
                                    featsstr += self.mtlmodel.idxtofeatstagset[i]

                    featsslicepreds.append(featsstr)

                allsbdpreds.extend(sbdpreds)
                allpospreds.extend(pospreds)
                allfeatspreds.extend(featsslicepreds)
                allwords.extend([s.split('\t')[0].strip() for s in slice[0]])

        allpospreds = [self.mtlmodel.postagsetcrf.get_item_for_index(p) for p in allpospreds]


        if sent_tag != 'auto':
            allsbdpreds = taggedsbdpreds

        return allsbdpreds,allpospreds, allfeatspreds, allwords

    def prepare_data_files(self):
        """
        Prepares the train and dev data files for training
        """
        def write_file(filename,mode='train'):

            if mode == 'dev':
                data = devdata
            else:
                data = traindata

            with open(filename,'w') as tr:
                length = -1
                for sent in data:
                    i = 0
                    while i < len(sent):
                        if isinstance(sent[i]['id'], tuple):
                            # fetch the super token tag
                            supertoken = 'B'
                            length = sent[i]['id'][-1] - sent[i]['id'][0]
                            start = sent[i]['id'][0]
                            i += 1
                            continue
                        elif length > 0 and supertoken in ('B','I'):
                            if sent[i]['id'] == start:
                                supertoken = 'B'
                            else:
                                supertoken = 'I'
                            length -=1
                        elif length == 0:
                            supertoken = 'E'
                            length = -1
                        elif length == -1:
                            supertoken = 'O'

                        if sent[i]['feats'] is not None:
                            feats = '|'.join(k + '=' + v for k,v in sent[i]['feats'].items())
                        else:
                            feats = ''

                        if sent[i]['id'] == 1:
                            tr.write(sent[i]['form'] + '\t' + sent[i]['upos'] + '\t' + 'B-SENT' + '\t' + supertoken + '\t' + feats + '\n')

                        else:
                            tr.write(sent[i]['form'] + '\t' + sent[i]['upos'] + '\t' + 'O' + '\t' + supertoken + '\t' + feats + '\n')

                        i += 1

        traindata = self.read_conllu()
        devdata = self.read_conllu(mode='dev')

        write_file(self.trainingdatafile,mode='train')
        write_file(self.devdatafile,mode='dev')

    def read_conllu(self,mode='train'):

        fields = tuple(
            list(conllu.parser.DEFAULT_FIELDS)
        )

        if mode == 'dev':
            file = self.devfile
        else:
            file = self.trainfile

        with open(file, "r", encoding="utf-8") as f:
            return conllu.parse(f.read(), fields=fields)

    def predict(self, xml_data,out_mode='conllu',sent_tag='auto',checkpointfile = None):

        def is_sgml_tag(line):
            return line.startswith("<") and line.endswith(">")

        def wrap_words(xml):
            output = []
            lines = xml.split("\n")
            for line in lines:
                if len(line)>0 and not (line.startswith("<") and line.endswith(">") and not line == "|"):
                    line = line.replace("|","\n")
                    line = "<❦♥>\n" + line + "\n</❦♥>"
                output.append(line)
            return "\n".join(output)

        def collapse_words(sgml):
            output = []
            buffer = []
            for line in sgml.split("\n"):
                if line in ['<❦♥>','</❦♥>'] or not is_sgml_tag(line):
                    buffer.append(line)
                else:
                    output.append(line)
                if line == "</❦♥>":
                    piped = "|".join(buffer)
                    if not (buffer[1] == "|" and len(buffer) == 3):  # Actual pipe as token
                        piped = piped.replace('|</❦♥>','</❦♥>').replace('<❦♥>|','<❦♥>')
                    output.append(piped)
                    buffer = []
            return "\n".join(output)

        def fix_malformed_sentences(sgml_list):
            """
            Fixing malformed SGML seems to boil down to two cases:

            (1) The sentence is interrupted by the close of a tag that opened before it. In this case,
                update the s boundaries so that we close and begin sentences at the close tag:

                                     <a>
                        <a>          ...
                        ...          <s>
                        <s>          ...
                        ...    ==>   </s>
                        </a>         </a>
                        ...          <s>
                        </s>         ...
                                     </s>

            (2) Some tag opened inside of the sentence and has remained unclosed at the time of sentence closure.
                In this case, we choose not to believe the sentence split, and merge the two sentences:

                        <s>
                        ...          <s>
                        <a>          ...
                        ...          <a>
                        </s>   ==>   ...
                        <s>          ...
                        ...          </a>
                        </a>         ...
                        ...          </s>
                        </s>
            """
            tag_opened = defaultdict(list)
            i = 0
            while i < len(sgml_list):
                line = sgml_list[i].strip()
                open_match = re.search(OPEN_SGML_ELT, line)
                close_match = re.search(CLOSE_SGML_ELT, line)
                if open_match:
                    tag_opened[open_match.groups()[0]].append(i)
                elif close_match:
                    tagname = close_match.groups()[0]
                    j = maximal_nontoken_span_end(sgml_list, i + 1)
                    mns = sgml_list[i:j]

                    # case 1: we've encountered a non-s closing tag. If...
                    if (
                            tagname != "s"  # the closing tag is not an s
                            and len(tag_opened["s"]) > 0  # and we're in a sentence
                            and len(tag_opened[tagname]) > 0
                            and len(tag_opened["s"]) > 0  # and the sentence opened after the tag
                            and tag_opened[tagname][-1] < tag_opened["s"][-1]
                            and "</s>" not in mns  # the sentence is not closed in the mns
                    ):
                        # end sentence here and move i back to the line we were looking at
                        sgml_list.insert(i, "</s>")
                        i += 1
                        # open a new sentence at the end of the mns and note that we are no longer in the sentence
                        sgml_list.insert(j + 1, "<s>")
                        tag_opened["s"].pop(-1)
                        # we have successfully closed this tag
                        tag_opened[tagname].pop(-1)
                    # case 2: s closing tag and there's some tag that opened inside of it that isn't closed in time
                    elif tagname == "s" and any(
                            e != "s" and f"</{e}>" not in mns
                            for e in [
                                e
                                for e in tag_opened.keys()
                                if len(tag_opened[e]) > 0 and len(tag_opened["s"]) > 0 and tag_opened[e][-1] >
                                   tag_opened["s"][-1]
                            ]
                    ):
                        # some non-s element opened within this sentence and has not been closed even in the mns
                        assert "<s>" in mns
                        sgml_list.pop(i)
                        i -= 1
                        sgml_list.pop(i + mns.index("<s>"))
                    else:
                        tag_opened[tagname].pop(-1)
                i += 1
            return sgml_list

        def maximal_nontoken_span_end(sgml_list, i):
            """Return j such that sgml_list[i:j] does not contain tokens
            and no element that is begun in the MNS is closed in it."""
            opened = []
            j = i
            while j < len(sgml_list):
                line = sgml_list[j]
                open_match = re.match(OPEN_SGML_ELT, line)
                close_match = re.match(CLOSE_SGML_ELT, line)
                if not (open_match or close_match):
                    break
                if open_match:
                    opened.append(open_match.groups()[0])
                if close_match and close_match.groups()[0] in opened:
                    break
                j += 1
            return j

        def get_bound_group_map(data):

            mapping = {}
            data = data.split("\n")
            # Ignore markup
            data = [u for u in data if not (u.startswith("<") and u.endswith(">"))]
            counter = 0
            for i, line in enumerate(data):
                super_token = line.replace("|", "") if line != "|" else "|"
                segs = line.split("|") if line != "|" else ["|"]
                for j, seg in enumerate(segs):
                    if len(segs) > 1 and j == 0:
                        mapping[counter] = (super_token, len(segs))
                        super_token = ""
                    counter += 1

            return mapping

        # These XML tags force a sentence break in the data, you can add more here:
        BLOCK_TAGS = ["sp", "head", "p", "figure", "caption", "list", "item"]
        BLOCK_TAGS += ["❦❦❦"]  # reserved tag for sentences in input based on newlines
        OPEN_SGML_ELT = re.compile(r"^<([^/ ]+)( .*)?>$")
        CLOSE_SGML_ELT = re.compile(r"^</([^/]+)>$")

        # Sometimes the tokenizer doesn't newline every elt
        xml_data = xml_data.replace("><", ">\n<")
        # Ad hoc fix for a tokenization error
        xml_data = xml_data.replace("°<", "°\n<")
        # Remove empty elements?
        # for elt in TAGS:
        #    xml_data = xml_data.replace(f"<{elt}>\n</{elt}>\n", "")
        xml_data = wrap_words(xml_data)

        # don't feed the sentencer our pos and lemma predictions, if we have them
        no_pos_lemma = re.sub(r"([^\n\t]*?)\t[^\n\t]*?\t[^\n\t]*?\n", r"\1\n", xml_data)
        split_indices, pos_tags, morphs, words = self.inference(no_pos_lemma,sent_tag=sent_tag,checkpointfile=checkpointfile)

        # for xml
        counter = 0
        splitted = []
        opened_sent = False
        para = True

        xml_data = xml_data.replace("<s>","<❦❦❦>").replace("</s>","</❦❦❦>")
        for line in xml_data.strip().split("\n"):
            if not is_sgml_tag(line):
                # Token
                if split_indices[counter] == 1 or para:
                    if opened_sent:
                        rev_counter = len(splitted) - 1
                        while is_sgml_tag(splitted[rev_counter]) and rev_counter > 0:
                            rev_counter -= 1
                        if rev_counter > 0:
                            splitted.insert(rev_counter + 1, "</s>")
                    splitted.append("<s>")
                    opened_sent = True
                    para = False
                counter += 1
            elif any(f"<{elt}>" in line for elt in BLOCK_TAGS) or any(
                f"</{elt}>" in line for elt in BLOCK_TAGS
            ):  # New block, force sentence split
                para = True
            splitted.append(line)

        if opened_sent:
            rev_counter = len(splitted) - 1
            while is_sgml_tag(splitted[rev_counter]):
                rev_counter -= 1
            splitted.insert(rev_counter + 1, "</s>")

        lines = "\n".join(splitted)
        lines = re.sub(r'</?❦❦❦>\n?','',lines)
        lines = reorder(lines, priorities=["s","❦♥"])
        lines = collapse_words(lines)

        # destroy any xml inside supertokens
        while re.search(r'(<❦♥>[^<>]*)<[^❦♥]+>',lines) is not None:
            lines = re.sub(r'(<❦♥>[^<>]*)<[^❦♥]+>([^<>]*</❦♥>)',r'\1\2',lines)

        # remove word and sent wrappers
        lines = re.sub(r'</?❦♥>','',lines)

        lines = reorder(lines)
        lines = fix_malformed_sentences(lines.split("\n"))
        lines = "\n".join(lines)
        lines = reorder(lines)

        # Split out the internal tags within MWT tokens, as these too get a POS tag
        data = lines.split("\n")
        retokenized = []
        for line in data:
            if line == "|":
                retokenized.append(line)
            else:
                retokenized.append("\n".join(line.split("|")))
        data = "\n".join(retokenized)

        """
        Now add the pos tags
        """
        bound_group_map = get_bound_group_map(data) if out_mode == "conllu" else None
        data = conllize(data, element="s", super_mapping=bound_group_map, attrs_as_comments=True)
        data = data.strip() + "\n"  # Ensure final new line for last sentence

        # add the pos tags to conllized file and remove the rows hyphenated MWT ID
        output = []
        tid = 1
        k = 0
        data = data.split('\n')
        for i in range(0,len(data)):
            if len(data[i].strip())==0:
                output.append("")
                tid = 1
            else:
                if "\t" in data[i]:
                    fields = data[i].split("\t")
                    if "." in fields[0] or "-" in fields[0]: continue
                    else:
                        fields = [str(tid), fields[1].strip(), "_", pos_tags[k].strip(), pos_tags[k].strip(), "_", "_", "_", "_", "_"]
                        output.append('\t'.join(fields))
                        tid += 1
                        k += 1

        assert k == len(pos_tags) # Fails means pos tags aren't aligned with tokens

        morphs = [m if m != '' else '_' for m in morphs]

        return "\n".join(output), lines, morphs, words

    def spans_score(self, gold_spans, system_spans):
        correct, gi, si = 0, 0, 0
        while gi < len(gold_spans) and si < len(system_spans):
            if system_spans[si].start < gold_spans[gi].start:
                si += 1
            elif gold_spans[gi].start < system_spans[si].start:
                gi += 1
            else:
                correct += gold_spans[gi].end == system_spans[si].end
                si += 1
                gi += 1

        return Score(len(gold_spans), len(system_spans), correct)



def main(): # testing only

    parser = argparse.ArgumentParser()
    parser.add_argument('--lr', type=float, default=1e-3)
    parser.add_argument('--seqlen', type=int, default=256)
    parser.add_argument('--trainbatch', type=int, default=32)
    parser.add_argument('--datatype', type=str, default='wiki')
    parser.add_argument('--sbdrnndim', type=int, default=256)
    parser.add_argument('--posrnndim', type=int, default=512)
    parser.add_argument('--morphrnndim', type=int, default=512)
    parser.add_argument('--sbdfflayerdim', type=int, default=256)
    parser.add_argument('--posfflayerdim', type=int, default=512)
    parser.add_argument('--morphfflayerdim', type=int, default=512)
    parser.add_argument('--posrnnbidirectional', type=bool, default=True)
    parser.add_argument('--sbdrnnbidirectional', type=bool, default=True)
    parser.add_argument('--morphrnnbidirectional', type=bool, default=True)
    parser.add_argument('--posrnnnumlayers', type=int, default=1)
    parser.add_argument('--sbdrnnnumlayers', type=int, default=1)
    parser.add_argument('--morphrnnnumlayers', type=int, default=1)
    parser.add_argument('--sbdencodertype', type=str, default='lstm')
    parser.add_argument('--posencodertype', type=str, default='lstm')
    parser.add_argument('--morphencodertype', type=str, default='lstm')
    parser.add_argument('--dropout', type=float, default=0.05)
    parser.add_argument('--worddropout', type=float, default=0.05)
    parser.add_argument('--lockeddropout', type=float, default=0.5)


    args = parser.parse_args()

    if args.datatype == 'wiki':
        trainfile = '../he_iahltwiki-ud-train.conllu'
        devfile = '../he_iahltwiki-ud-dev.conllu'
    else:
        devfile = '../he_htb-ud-dev.conllu'
        trainfile = '../he_htb-ud-train.conllu'

    tagger = Tagger(trainflag=True, trainfile=trainfile, devfile=devfile, sbdrnndim=args.sbdrnndim,
                    sbdfflayerdim=args.sbdfflayerdim,
                    posrnndim=args.posrnndim, posfflayerdim=args.posfflayerdim,
                    sbdrnnbidirectional=args.sbdrnnbidirectional,
                    posrnnbidirectional=args.posrnnbidirectional, sbdrnnnumlayers=args.sbdrnnnumlayers,
                    posrnnnumlayers=args.posrnnnumlayers, sbdencodertype=args.sbdencodertype,
                    posencodertype=args.posencodertype,
                    morphrnnbidirectional=args.morphrnnbidirectional, morphrnndim=args.morphrnndim,
                    morphfflayerdim=args.morphfflayerdim, morphencodertype=args.morphencodertype,
                    morphrnnnumlayers=args.morphrnnnumlayers,
                    learningrate=args.lr, batchsize=args.trainbatch, sequencelength=args.seqlen,
                    dropout=args.dropout, wordropout=args.worddropout, lockeddropout=args.lockeddropout,
                    datatype=args.datatype)

    tagger.prepare_data_files()
    tagger.train()



if __name__ == "__main__":
    main()



from setuptools import setup, find_packages

setup(
  name = 'hebpipe',
  packages = find_packages(),
  version = '3.0.0.0',
  description = 'A pipeline for Hebrew NLP',
  author = 'Amir Zeldes',
  author_email = 'amir.zeldes@georgetown.edu',
  package_data = {'':['README.md','LICENSE.md','requirements.txt'],'hebpipe':['lib/*','data/*','bin/*','models/models_go_here.txt','models/stanza/stanza_models_here.txt']},
   install_requires=['requests','numpy','transformers==3.5.1','torch==1.6.0','pandas','scipy','joblib','xgboost==0.81','rftokenizer','depedit','xmltodict', 'diaparser==1.1.2','flair==0.6.1','stanza','conllu'],
  url = 'https://github.com/amir-zeldes/HebPipe',
  license='Apache License, Version 2.0',
  download_url = 'https://github.com/amir-zeldes/HebPipe/releases/tag/v3.0.0.0',
  keywords = ['NLP', 'Hebrew', 'segmentation', 'tokenization', 'tagging', 'parsing','morphology','POS','lemmatization'],
  classifiers = ['Programming Language :: Python',
'Programming Language :: Python :: 2',
'Programming Language :: Python :: 3',
'License :: OSI Approved :: Apache Software License',
'Operating System :: OS Independent'],
)


from sklearn.metrics import accuracy_score, classification_report, f1_score
from collections import defaultdict

tags = {"NOUN","VERB","ADJ","ADV","PROPN"}
lex = open("C:\\Uni\\Hebrew\\HebPipe.git\\hebpipe\\data\\heb.lex",encoding="utf8").read().strip().split("\n")
lex = [l for l in lex if "\t" in l]
lex = [l for l in lex if l.split("\t")[1] in tags]
lex = {l.split("\t")[0] + "\t" + l.split("\t")[1]:l.split("\t")[2] for l in lex if "\t" in l}

def clean_final(text):
    finals = {"פ": "ף", "כ": "ך", "מ": "ם", "נ": "ן", "צ": "ץ"}
    if text[-1] in finals:
        text = text[:-1] + finals[text[-1]]
    return text


def post_process(word, pos, lemma, morph):
    if "Plur" in morph and word == lemma and pos in ["NOUN", "ADJ"] and (word.endswith("ים") or word.endswith("ות")):
        lemma = lemma[:-2]
        if word.endswith("ות"):
            lemma += "ה"
        lemma = clean_final(lemma)
    #if "Fem" in morph and pos == "ADJ" and word.endswith("ית") and lemma == word:
    #    lemma = lemma[:-1]
    if "Fem" in morph and word == lemma:
        if word + "\t" + pos in lex:
            pass
            #lemma = lex[word+"\t"+pos]
    if word == lemma:
        if word + "\t" + pos in lex:
            if pos == "VERB" and "Fut" in morph:
                lemma = lex[word + "\t" + pos]
            if pos == "VERB" and "Pres" in morph:
                lemma = lex[word + "\t" + pos]
            if pos == "VERB" and "Part" in morph:
                lemma = lex[word + "\t" + pos]
            if pos in ["NOUN","ADJ"] and "Plur" in morph:
                lemma = lex[word + "\t" + pos]

    return lemma


def get_annos(morph):
    if morph == "_":
        return {}
    else:
        annos = morph.split("|")
        output = {}
        for anno in annos:
            k, v = anno.split("=")
            output[k] = v
    return output

gold_file = "C:\\Uni\\Corpora\\Hebrew\\IAHLT_HTB\\he_htb-ud-test.conllu"
pred_file = "he_htb-ud-test_pred.conllu"
#pred_file = "he_htb-ud-test_goldpos_pred.conllu"
#pred_file = "he_htb-ud-test_marmot_pred.conllu"

preds = []

metric = "acc"#"f1"
target = "lemma"#"morph"

print("Base: 0.9653150134048257")

for line in open(pred_file).read().split("\n"):
    if "\t" in line:
        fields = line.split("\t")
        if "-" in fields[0]:
            continue
        fields[2] = post_process(fields[1],fields[4],fields[2],fields[5])
        if target == "pos":
            preds.append(fields[4])
        elif target == "lemma":
            preds.append(fields[2])
        else:
            preds.append(fields[5])

gold = []
for line in open(gold_file).read().split("\n"):
    if "\t" in line:
        fields = line.split("\t")
        if "-" in fields[0]:
            continue
        if target == "pos":
            gold.append(fields[4])
        elif target == "lemma":
            gold.append(fields[2])
        else:
            gold.append(fields[5])
tok_count = len(gold)

if target == "morph":
    all_preds = defaultdict(lambda :defaultdict(lambda :"_"))
    all_golds = defaultdict(lambda :defaultdict(lambda :"_"))
    all_annos = set()
    for i, g in enumerate(gold):
        anno_kv = get_annos(g)
        for anno, val in anno_kv.items():
            all_golds[i][anno] = val
            all_annos.add(anno)
        anno_kv = get_annos(preds[i])
        for anno, val in anno_kv.items():
            all_preds[i][anno] = val
            all_annos.add(anno)

    for anno in sorted(list(all_annos)):
        if metric=="acc":
            golds = [all_golds[i][anno] for i in range(tok_count)]
            preds = [all_preds[i][anno] for i in range(tok_count)]
            print(anno+"\t",end="")
            print(accuracy_score(golds, preds))
        else:
            golds = [all_golds[i][anno] for i in range(tok_count) if not (all_golds[i][anno] == "_" and all_preds[i][anno] == "_")]
            preds = [all_preds[i][anno] for i in range(tok_count) if not (all_golds[i][anno] == "_" and all_preds[i][anno] == "_")]
            labs = list(set([all_preds[i][anno] for i in all_preds]))
            print(anno+"\t",end="")
            print(f1_score(golds, preds,average="micro",labels=labs))


else:

    print(accuracy_score(gold,preds))
    errs = [str((g,preds[i])) for i, g in enumerate(gold) if preds[i] != g]
    print("\n".join(errs))
    print(len(errs))
    if target != "lemma":
        print(classification_report(gold,preds))


import re, operator

class CorefRule:
	def __init__(self,rule_string, rule_num):
		if not 3 <= rule_string.count(";") <= 5:
			raise Exception("coref rule does not contain 3-5 semicolons: " + rule_string)
		parts = rule_string.split(";")
		self.ana_spec, self.ante_spec, self.max_distance, self.propagation = parts[0:4]
		if len(parts) > 4:
			self.clf_name = parts[4]
		else:
			self.clf_name = "_default_"
		if len(parts) == 6:
			self.thresh = float(parts[5])
		else:
			self.thresh = None
		self.max_distance = int(self.max_distance)
		self.ana_constraints = []
		self.ante_constraints = []
		for item in self.ana_spec.split("&"):
			self.ana_constraints.append(ConstraintMatcher(item))
		for item in self.ante_spec.split("&"):
			self.ante_constraints.append(ConstraintMatcher(item))
		# Make sure that group failure criteria are first to be checked
		self.ante_constraints.sort(key=lambda x: x.group_failure, reverse=True)
		self.rule_num = rule_num

	def __repr__(self):
		return self.ana_spec + " -> " + self.ante_spec + " (" + str(self.max_distance) + ", " + self.propagation + ", "+ self.clf_name + ")"


class ConstraintMatcher:
	def __init__(self,constraint):
		self.group_failure = False
		self.negative = operator.truth
		self.match_type = "exact"
		self.value = ""
		self.key = ""
		self.compiled_re = None
		self.props = {"form", "text", "agree", "entity", "subclass", "cardinality","text_lower","lemma","pos","func","quoted","mood","speaker"}

		if constraint.endswith("*"):
			self.group_failure = True
			constraint = constraint[0:-1]

		if "=" in constraint:  # Key-value constrains
			key, value = constraint.split("=")
			if key[-1] == "!":
				self.negative = operator.not_
				key = key[0:-1]

			if value.startswith('"') and value.endswith('"'):
				value = value[1:-1]
				self.match_type = "exact"
			elif value.startswith("/") and value.endswith("/"):
				value = value[1:-1]
				if re.escape(value) == value:  # Does not contain special characters but is medial regex
					self.match_type = "exact"
				elif re.escape(value[1:-1]) == value and value.startswith("^") and value.endswith("$"):
					# Regex only supplies anchors, treat as exact match
					value = value[1:-1]
					self.match_type = "exact"
				elif re.escape(value[1:]) == value and value.startswith("^"):
					# Regex only supplies initial anchor, treat as startswith
					value = value[1:]
					self.match_type = "startswith"
				elif re.escape(value[:-1]) == value and value.endswith("$"):
					# Regex only supplies initial anchor, treat as startswith
					value = value[:-1]
					self.match_type = "endswith"
				else:
					self.match_type = "regex"
					self.compiled_re = re.compile(value)
			elif value.lower() == "true":
				self.match_type = "bool"
				self.value = True
			elif value.lower() == "false":
				self.match_type = "bool"
				self.value = False
			elif value.startswith("$"):
				# Antecedent-based spec, can't precompile value matcher
				self.match_type = "dollar"
			else:  # String literal without regex
				self.match_type = "exact"
			if self.match_type != "bool":
				self.value = value
			self.key = key
		elif constraint == "none" or constraint.startswith("any") or constraint.startswith("look") or constraint.startswith("take"):
			# This is a 'none' matcher or processing instruction, matches anything
			self.match_type = "none"
		elif "sameparent" in constraint:  # same or !same style constraint - port to $ for backwards compatibility
			if constraint[0] == "!":
				self.negative = operator.not_
			self.match_type = "dollar"
			self.key = "parent"
			self.value = "$1"
		elif "samespeaker" in constraint:  # same or !same style constraint - port to $ for backwards compatibility
			if constraint[0] == "!":
				self.negative = operator.not_
			self.match_type = "dollar"
			self.key = "speaker"
			self.value = "$1"
		elif constraint.startswith("last["):
			self.match_type = "exact"
			self.key = "LAST"
			self.value = constraint[constraint.find("[")+1:-1]

	def __repr__(self):
		if self.negative == operator.truth:
			op = ""
		else:
			op = "!"
		return self.key + " " + op + self.match_type + " '" + self.value + "'"

	def match(self,mark,lex,anaphor=None):
		test_val = ""
		op = self.negative

		if self.match_type == "none":
			return True
		elif self.match_type == "dollar":
			if self.key in self.props:
				self.value = str(getattr(anaphor, self.key))
				test_val = str(getattr(mark, self.key))
			elif self.key == "head":
				return op(anaphor.head.id == mark.head.head)
			elif self.key == "child":
				return op(anaphor.head.head == mark.head.id)
			elif self.key == "hasa":
				return op(anaphor.head.head_text in lex.hasa[mark.lemma])
			elif self.key == "parent":
				if mark.head.head == "0":  # Root token, by definition not same parent as another token
					retval = op(False)
				elif mark.sentence.sent_num != anaphor.sentence.sent_num:
					retval = op(False)
				else:
					retval = op(anaphor.head.head == mark.head.head)
				if retval is False and self.group_failure and anaphor is not None:
					mark.non_antecdent_groups.add(anaphor.group)
					anaphor.non_antecdent_groups.add(mark.group)
				return retval
			elif self.key == "has_child_func":
				raise Exception("coref rule 'has_child_func=$' : $ identity not implemented for has_child_func")
			elif self.key == "mod":
				mods = getattr(anaphor.head, "modifiers")
				found_mod = False
				for mod1 in mark.head.modifiers:
					for mod2 in mods:
						if mod1.lemma == mod2.lemma and lex.filters["det_func"].match(mod1.func) is None and \
						lex.filters["det_func"].match(mod2.func) is None:
							found_mod = True
				if not found_mod:
					if self.group_failure and anaphor is not None:
						mark.non_antecdent_groups.add(anaphor.group)
					return False
				else:
					return True
		else:
			if self.key in self.props:
				if self.match_type == "bool":
					test_val = getattr(mark, self.key)
				else:
					test_val = str(getattr(mark, self.key))
			elif self.key == "LAST":
				if self.value in lex.last:
					return op(lex.last[self.value].entity==mark.entity)
				else:
					return False
			elif self.key == "has_child_func":
				test_val = mark.child_func_string
				self.match_type = "substring"
				if self.value[0] != ";":
					self.value = ";"+self.value+";"
			elif self.key == "mod":
				mods = [self.value]
				found_mod = False
				for mod1 in mark.head.modifiers:
					for mod2 in mods:
						# Note that mod2 is just a string - not a ParsedToken
						if mod1.lemma == mod2 and lex.filters["det_func"].match(mod1.func) is None:
							found_mod = True
				if not found_mod:
					if self.group_failure and anaphor is not None:
						mark.non_antecdent_groups.add(anaphor.group)
					return False
				else:
					return True
			elif self.key == "head":
				raise Exception("coref rule 'head=VAL' : value match not implemented for head")
			elif self.key == "child":
				raise Exception("coref rule 'child=VAL' : value match not implemented for child")

		if self.match_type == "exact":
			retval = op(test_val == self.value)
		elif self.match_type == "substring":
			retval = op(self.value in test_val)
		elif self.match_type == "regex":
			retval = op(self.compiled_re.search(test_val) is not None)
		elif self.match_type == "startswith":
			retval = op(test_val.startswith(self.value))
		elif self.match_type == "endswith":
			retval = op(test_val.endswith(self.value))
		elif self.match_type == "dollar":
			retval = op(test_val == self.value)
		elif self.match_type == "bool":
			retval = op(test_val == self.value)

		if retval is False and self.group_failure and anaphor is not None:
			mark.non_antecdent_groups.add(anaphor.group)

		return retval




from collections import defaultdict
import re, sys
import xml.dom.minidom


def printToFile(content, filename):
	firstname = filename.split(".")[0]
	newname = firstname + "_cl.coref"
	f = open(newname, 'w')
	f.write(content)
	f.close()


def removeEmptyCat(filename):
	f = open(filename, 'r')
	raw = f.read()
	insBefore = addNewlineBefore(raw)
	insAfter = addNewlineAfter(insBefore)

	tokens = re.split(' |\n', insAfter)
	new = ""
	for tok in tokens:
		star = re.search(r"(^|[^\\])\*", tok)
		if (not star and tok != "0") or tok.count("*") > 2 or (
				tok.count("*") < 3 and re.search(r'\*[a-z][a-z][a-z][a-z]+', tok) is not None):
			new = new + " " + tok

	return new


def addNewlineBefore(text):
	tokens = text.split("<")
	new = ""
	for tok in tokens[1:]:
		new = new + "\n" + "<" + tok
	return new


def addNewlineAfter(text):
	# same approach to add a \n after >
	tokens = text.split(">")
	new = ""
	for tok in tokens:
		new = new + tok + ">" + "\n"
	return new


def mainPreprocess(filen):
	"""preprocess OntoNotes gold files by cleaning up the empty categories and more"""
	# filen=sys.argv[1]
	print "Processing " + filen
	noemp = removeEmptyCat(filen)
	noempi = noemp[:-2]
	a = xml.dom.minidom.parseString(noempi)
	# pretty_xml_as_string = a.toprettyxml()
	return a


class OntoMark:
	def __init__(self, start, end, text, group):
		self.start = start
		self.end = end
		self.text = text.strip()
		self.group = group
		self.span = str(start) + "-" + str(end)

	def __repr__(self):
		return "(" + str(self.start) + "," + str(self.end) + ") group:" + str(self.group) + " text:" + self.text


class Oracle():

	def __init__(self):
		self.correct_pairs = []
		self.incorrect_triples = [] # Triples of anaphor - wrong_antecedent - correct antecedent

	def process_onto_mark(self, node, start):
		"""
		:param node: an Element node representing a markable
		:param start: the token number of the first tok
		:return: a list whose first item if number of tokens in this element, and following items are OntoMark objects
				representing this markable and any embedded within it (found recursively)
		"""
		project_appos_wrapper = False
		group = node.getAttribute("ID")
		if len(node.childNodes) == 1: # node is just one markable, child should be a Text node
			text = node.firstChild.data
			split_text = text.split()
			end = start + len(split_text) - 1

			return [len(split_text), OntoMark(start, end, text, group)]
		else:
			text = ""
			token_counter = 0
			child_marks = []
			for child in node.childNodes:
				if child.nodeType == 3: # Text
					text += child.data
					token_counter = len(text.split())
				elif child.nodeType == 1:
					if project_appos_wrapper:
						if child.getAttribute("TYPE") == "APPOS": # fix appos wrapper issue
							child.setAttribute("ID", node.getAttribute("ID"))
					process_child = self.process_onto_mark(child, start + token_counter)
					text += process_child[1].text
					child_marks += process_child[1:]
					token_counter = len(text.split())
			return [token_counter, OntoMark(start, start + token_counter - 1, text, group)] + child_marks



	def get_gold_data(self, docname, coref_dir="C:\\Uni\\CL\\xrenner\\resources\\anto_emma\\"):

		xml_filename = coref_dir + docname + ".coref"

		# preprocess and open xml file
		xml_file = mainPreprocess(xml_filename)
		xml_list = xml_file.childNodes[0].childNodes[1].childNodes

		self.group_dict = defaultdict(list)

		self.mark_dict = {}

		# get a list of  onto markables to cross-check with later
		xml_tok_counter = 1
		xml_marks = []
		for node in xml_list:
			if xml_tok_counter > 45:
				pass
			if node.nodeType == 3: # text
				s = node.data.split()
				xml_tok_counter += len(s)
			elif node.nodeType == 1: # COREF element
				processed_mark = self.process_onto_mark(node, xml_tok_counter)
				xml_tok_counter += processed_mark[0]
				new_marks = processed_mark[1:]
				for new_mark in new_marks:
					xml_marks.append(new_mark)
					self.group_dict[new_mark.group].append(new_mark)
					self.mark_dict[str(new_mark.start) + "-" + str(new_mark.end)] = new_mark


	def eval_gold(self, xrenner):
		self.get_gold_data(xrenner.docname)
		out_correct = open("c:\\uni\\cl\\xrenner\\resources\\oracle_correct.tab",'w')
		out_incorrect = open("c:\\uni\\cl\\xrenner\\resources\\oracle_incorrect.tab",'w')

		for mark in xrenner.markables:
			mark_span = str(mark.start) + "-" + str(mark.end)
			if mark.antecedent != "none":
				antecedent = mark.antecedent
				antecedent_span = str(antecedent.start) + "-" + str(antecedent.end)
			else:
				antecedent_span = ""

			if mark_span in self.mark_dict:
				onto_mark = self.mark_dict[mark_span]
				mark_gold_group = onto_mark.group
				mark_group_position = self.group_dict[mark_gold_group].index(onto_mark)
				if mark_group_position == 0:
					onto_antecedent_span = ""
					onto_antecedent_text = "FIRST"
				else:
					onto_antecedent_span = self.group_dict[mark_gold_group][mark_group_position-1].span
					onto_antecedent_text = self.mark_dict[onto_antecedent_span].text

				if antecedent_span == onto_antecedent_span == "":
					# Correct antecedent
					self.correct_pairs.append(mark.text + "\tFIRST")
				elif antecedent_span == onto_antecedent_span:
					# Correct FIRST
					self.correct_pairs.append(mark.text + "\t" + mark.antecedent.text)
				elif antecedent_span not in self.mark_dict or onto_antecedent_span != antecedent_span:
					# False antecedent
					if mark.antecedent == "none":
						self.incorrect_triples.append(mark.text + "\t" + mark.head.text + "\tFIRST\tFIRST\t" + onto_antecedent_text)
					else:
						self.incorrect_triples.append(mark.text + "\t" + mark.head.text + "\t" + mark.antecedent.text + "\t" + mark.antecedent.head.text + "\t" + onto_antecedent_text)
			else:
				# False positive
				if mark.antecedent == "none":
					#self.incorrect_triples.append(mark.text + "\tFIRST\tSINGLETON")
					pass
				else:
					if mark.text == "I":
						pass
					self.incorrect_triples.append(mark.text + "\t" + mark.head.text + "\t" + mark.antecedent.text + "\t" + mark.antecedent.head.text + "\tSINGLETON")

		for item in self.correct_pairs:
			out_correct.write(item + "\n")

		for item in self.incorrect_triples:
			out_incorrect.write(item + "\n")



#!/usr/bin/python
# -*- coding: utf-8 -*-

import re, sys, io, os, platform
import tempfile
import subprocess
from glob import glob
from diaparser.parsers import Parser
from stanza.utils.conll import CoNLL
from depedit import DepEdit
import stanza
from stanza.models.common.doc import Document
import torch

from rftokenizer import RFTokenizer
try:  # Module usage
    from .lib.xrenner import Xrenner
    from .lib._version import __version__
    from .lib.tt2conll import conllize
    from .lib.append_column import inject_col
    from .lib.sent_split import toks_to_sents
    from .lib.whitespace_tokenize import add_space_after, tokenize as whitespace_tokenize
    from .lib.mtlmodel import Tagger
except ImportError:  # direct script usage
    from lib.xrenner import Xrenner
    from lib._version import __version__
    from lib.tt2conll import conllize
    from lib.append_column import inject_col
    from lib.sent_split import toks_to_sents
    from lib.whitespace_tokenize import add_space_after, tokenize as whitespace_tokenize
    from lib.mtlmodel import Tagger

PY3 = sys.version_info[0] > 2

if PY3:
    from urllib.request import urlretrieve
else:
    from urllib import urlretrieve

inp = input if PY3 else raw_input

script_dir = os.path.dirname(os.path.realpath(__file__))
lib_dir = script_dir + os.sep + "lib" + os.sep
bin_dir = script_dir + os.sep + "bin" + os.sep
data_dir = script_dir + os.sep + "data" + os.sep
model_dir = script_dir + os.sep + "models" + os.sep
marmot_path = bin_dir + "Marmot" + os.sep

KNOWN_PUNCT = {'’','“','”'}  # Hardwired tokens to tag as punctutation (unicode glyphs not in training data)
morph_deped = DepEdit(config_file=lib_dir+"partial_morph.ini")
tags = {"NOUN", "VERB", "ADJ", "ADV", "PROPN"}
lex_data = open(data_dir + "heb.lemma", encoding="utf8").readlines()
lex = {}
for l in lex_data:
    word, tag, lemma = l[:-1].split("\t")
    if tag in tags:
        lex["\t".join([word,tag])] = lemma

binyan_data = open(data_dir + "heb.binyan", encoding="utf8").readlines()
binyan_lookup = {}
binyan_lemma_lookup = {}
for l in binyan_data:
    word, lemma, binyan = l[:-1].split("\t")
    binyan_lookup[(word,lemma)] = binyan
    binyan_lemma_lookup[lemma] = binyan


def init_lemmatizer(cpu=False, no_post_process=False):
    lemmatizer = stanza.Pipeline("he", package="htb", processors="lemma", tokenize_no_ssplit=True,
                         tokenize_pretokenized=True,
                          lemma_pretagged=True,
                          depparse_pretagged=True,
                          use_gpu=not cpu,
                          lemma_model_path=model_dir + "stanza" + os.sep + "he_lemmatizer.pt")
    lemmatizer.do_post_process = False if no_post_process else True
    return lemmatizer


def log_tasks(opts):
    sys.stderr.write("\nRunning tasks:\n" +"="*20 + "\n")
    if opts.sent not in ["auto","none"]:
        sys.stderr.write("o Splitting sentences based on tag: "+opts.sent+"\n")
    elif opts.sent == "auto":
        if opts.punct_sentencer:
            sys.stderr.write("o Automatic sentence splitting (punctuation based)\n")
        else:
            sys.stderr.write("o Automatic sentence splitting (neural)\n")
    if opts.whitespace:
        sys.stderr.write("o Whitespace tokenization\n")
    if opts.tokenize:
        sys.stderr.write("o Morphological segmentation\n")
    if opts.posmorph:
        sys.stderr.write("o POS and Morphological tagging\n")
    if opts.lemma:
        sys.stderr.write("o Lemmatization\n")
    if opts.dependencies:
        sys.stderr.write("o Dependency parsing\n")
    if opts.entities:
        sys.stderr.write("o Entity recognition\n")
    if opts.coref:
        sys.stderr.write("o Coreference resolution\n")

    sys.stderr.write("\n")


def diagnose_opts(opts):
    if not opts.posmorph and not opts.whitespace and not opts.tokenize and not opts.lemma \
        and not opts.dependencies and not opts.entities and not opts.coref:
        if not opts.quiet:
            sys.stderr.write("! You selected no processing options\n")
            sys.stderr.write("! Assuming you want all processing steps\n")
        opts.whitespace = True
        opts.tokenize = True
        opts.posmorph = True
        opts.lemma = True
        opts.dependencies = True
        opts.entities = True
        opts.coref = True

    added = []
    trigger = ""
    if opts.dependencies:
        trigger = "depenedencies"
        if not opts.posmorph:
            added.append("posmorph")
            opts.posmorph = True
        if not opts.lemma:
            added.append("lemma")
            opts.lemma = True

    if len(added)>0:
        sys.stderr.write("! You selected "+trigger+"\n")
        sys.stderr.write("! Turning on options: "+",".join(added) +"\n")

    added = []
    if opts.whitespace:
        trigger = "whitespace tokenization"
        if not opts.tokenize:
            added.append("tokenize")
            opts.tokenize = True
    if len(added)>0:
        sys.stderr.write("! You selected "+trigger+"\n")
        sys.stderr.write("! Turning on options: "+",".join(added) +"\n")
    return opts


def groupify(output,anno):
    groups = ""
    current_group = ""
    for line in output.split("\n"):
        if " "+anno+"=" in line:
            current_group += re.search(anno + r'="([^"]*)"',line).group(1)
        if line.startswith("</") and "_group" in line:
            groups += current_group +"\n"
            current_group = ""

    return groups


def get_bound_group_map(data):

    mapping = {}
    data = data.split("\n")
    # Ignore markup
    data = [u for u in data if not (u.startswith("<") and u.endswith(">"))]
    counter = 0
    for i, line in enumerate(data):
        super_token = line.replace("|","") if line != "|" else "|"
        segs = line.split("|") if line != "|" else ["|"]
        for j, seg in enumerate(segs):
            if len(segs)>1 and j == 0:
                mapping[counter] = (super_token,len(segs))
                super_token = ""
            counter += 1

    return mapping


def remove_nesting_attr(data, nester, nested, attr="xml:lang"):
    """
    Removes attribute on nesting element if a nested element includes it

    :param data: SGML input
    :param nester: nesting tag, e.g. "norm"
    :param nested: nested tag, e.g. "morph"
    :param attr: attribute, e.g. "lang"
    :return: cleaned SGML
    """

    if attr not in data:
        return data
    flagged = []
    in_attr_nester = False
    last_nester = -1
    lines = data.split("\n")
    for i, line in enumerate(lines):
        if nester + "=" in line and attr+"=" in line:
            in_attr_nester = True
            last_nester = i
        if "</" + nester + ">" in line:
            in_attr_nester = False
        if nested in line and attr+"=" in line and in_attr_nester and last_nester > -1:
            flagged.append(last_nester)
            in_attr_nester = False
    for i in flagged:
        lines[i] = re.sub(' '+attr+'="[^"]+"','',lines[i])
    return "\n".join(lines)


def tok_from_norm(data):
    """
    Takes TT-SGML, extracts norm attribute, and replaces existing tokens with norm values while retaining SGML tags.
    Used to feed parser norms while retaining SGML sentence separators.

    :param data: TTSGML with <norm norm=...> and raw tokens to replace
    :return: TTSGML with tags preserved and tokens replace by norm attribute values
    """

    outdata = []
    norm = ""
    for line in data.replace("\r","").split("\n"):
        if line.startswith("<"):
            m = re.search(r'norm="([^"]*)"',line)
            if m is not None:
                norm = m.group(1)
            outdata.append(line)
        else:
            if norm != "":
                outdata.append(norm)
                norm=""
    return "\n".join(outdata) + "\n"


def read_attributes(input,attribute_name):
    out_stream =""
    for line in input.split('\n'):
        if attribute_name + '="' in line:
            m = re.search(attribute_name+r'="([^"]*)"',line)
            if m is None:
                print("ERR: cant find " + attribute_name + " in line: " + line)
                attribute_value = ""
            else:
                attribute_value = m.group(1)
            if len(attribute_value)==0:
                attribute_value = "_warn:empty_"+attribute_name+"_"
            out_stream += attribute_value +"\n"
    return out_stream


def merge_into_tag(tag_to_kill, tag_to_merge_into,stream):
    vals = []
    cleaned_stream = ""
    for line in stream.split("\n"):
        if " "+tag_to_kill + "=" in line:
            val = re.search(" " + tag_to_kill+'="([^"]*)"',line).group(1)
            vals.append(val)
        elif "</" + tag_to_kill + ">" in line:
            pass
        else:
            cleaned_stream += line + "\n"
    injected = inject(tag_to_kill,"\n".join(vals).strip(),tag_to_merge_into,cleaned_stream)
    return injected


def inject_conllu(source_conllu, target_conllu, columns=None):
    if columns is None:
        columns = [6,7]
    lines = source_conllu.split("\n")
    insertions = []
    for line in lines:
        if "\t" in line:
            fields = line.split("\t")
            insertion = []
            for colnum in columns:
                insertion.append(fields[colnum])
            insertions.append(insertion)
    output = []
    toknum = 0
    for line in target_conllu.split("\n"):
        if "\t" in line:
            fields = line.split("\t")
            if "-" not in fields[0]:
                insertion = insertions[toknum]
                for i, val in enumerate(insertion):
                    fields[columns[i]] = val
                line = "\t".join(fields)
                toknum += 1
        output.append(line)
    return "\n".join(output)


def get_col(data, colnum):
    if not isinstance(data,list):
        data = data.split("\n")

    splits = [row.split("\t") for row in data if "\t" in row]
    return [r[colnum] for r in splits]


def exec_via_temp(input_text, command_params, workdir="", outfile=False):
    temp = tempfile.NamedTemporaryFile(delete=False)
    if outfile:
        temp2 = tempfile.NamedTemporaryFile(delete=False)
    output = ""
    try:
        temp.write(input_text.encode("utf8"))
        temp.close()

        if outfile:
            command_params = [x if 'tempfilename2' not in x else x.replace("tempfilename2",temp2.name) for x in command_params]
        command_params = [x if 'tempfilename' not in x else x.replace("tempfilename",temp.name) for x in command_params]
        if workdir == "":
            proc = subprocess.Popen(command_params, stdout=subprocess.PIPE,stdin=subprocess.PIPE,stderr=subprocess.PIPE)
            (stdout, stderr) = proc.communicate()
        else:
            proc = subprocess.Popen(command_params, stdout=subprocess.PIPE,stdin=subprocess.PIPE,stderr=subprocess.PIPE,cwd=workdir)
            (stdout, stderr) = proc.communicate()
        if outfile:
            if PY3:
                output = io.open(temp2.name,encoding="utf8").read()
            else:
                output = open(temp2.name).read()
            temp2.close()
            os.remove(temp2.name)
        else:
            output = stdout
        #print(stderr)
        proc.terminate()
    except Exception as e:
        print(e)
    finally:
        os.remove(temp.name)
        return output


def exec_via_temp_old(input_text, command_params, workdir=""):
    temp = tempfile.NamedTemporaryFile(delete=False)
    exec_out = ""
    try:
        if PY3:
            temp.write(input_text.encode("utf8"))
        else:
            temp.write(input_text.encode("utf8"))
        temp.close()

        command_params = [x if x != 'tempfilename' else temp.name for x in command_params]
        if workdir == "":
            proc = subprocess.Popen(command_params, stdout=subprocess.PIPE,stdin=subprocess.PIPE,stderr=subprocess.PIPE)
            (stdout, stderr) = proc.communicate()
        else:
            proc = subprocess.Popen(command_params, stdout=subprocess.PIPE,stdin=subprocess.PIPE,stderr=subprocess.PIPE,cwd=workdir)
            (stdout, stderr) = proc.communicate()

        exec_out = stdout
    except Exception as e:
        print(e)
    finally:
        os.remove(temp.name)
        #if PY3:
        exec_out = exec_out.decode("utf8")
        return exec_out


def get_origs(data):
    origs = []
    current = ""
    for line in data.split("\n"):
        if "</norm>" in line:
            origs.append(current)
            current = ""
        if not line.startswith("<"):  # Token line
            current += line

    return "\n".join(origs)


def inject(attribute_name, contents, at_attribute,into_stream,replace=True):
    insertions = contents.split('\n')
    injected = ""
    i=0
    for line in into_stream.split("\n"):
        if at_attribute + "=" in line:
            if i >= len(insertions):
                raise Exception("Error out of bounds at element " + str(i) + " in document beginning " + into_stream[:1000])
            if len(insertions[i])>0:
                if at_attribute == attribute_name:  # Replace old value of attribute with new one
                    line = re.sub(attribute_name+'="[^"]+"',attribute_name+'="'+insertions[i]+'"',line)
                else:  # Place before specific at_attribute
                    if replace or " " + attribute_name + "=" not in line:
                        line = re.sub(at_attribute+"=",attribute_name+'="'+insertions[i]+'" '+at_attribute+"=",line)
            i += 1
        injected += line + "\n"
    return injected


def extract_conll(conll_string):
    conll_string = conll_string.replace("\r","").strip()
    sentences = conll_string.split("\n\n")
    ids = ""
    funcs = ""
    parents = ""
    id_counter = 0
    offset = 0
    for sentence in sentences:
        tokens = sentence.split("\n")
        for token in tokens:
            if "\t" in token:
                id_counter +=1
                ids += "u"+ str(id_counter) + "\n"
                cols = token.split("\t")
                funcs += cols[7].replace("ROOT","root") +"\n"
                if cols[6] == "0":
                    parents += "#u0\n"
                else:
                    parents += "#u" + str(int(cols[6])+offset)+"\n"
        offset = id_counter
    return ids, funcs, parents


def inject_tags(in_sgml,insertion_specs,around_tag="norm",inserted_tag="mwe"):
    """

    :param in_sgml: input SGML stream including tags to surround with new tags
    :param insertion_specs: list of triples (start, end, value)
    :param around_tag: tag of span to surround by insertion
    :return: modified SGML stream
    """
    if len(insertion_specs) == 0:
        return in_sgml

    counter = -1
    next_insert = insertion_specs[0]
    outlines = []
    for line in in_sgml.split("\n"):
        if line.startswith("<" + around_tag + " "):
            counter += 1
            if next_insert[0] == counter:  # beginning of a span
                outlines.append("<" + inserted_tag + " " + inserted_tag + '="' + next_insert[2] + '">')
        outlines.append(line)
        if line.startswith("</" + around_tag + ">"):
            if next_insert[1] == counter:  # end of a span
                outlines.append("</" + inserted_tag + ">")

    return "\n".join(outlines)


def lemmatize(lemmatizer, conllu, morphs):
    def clean_final(text):
        finals = {"פ":"ף","כ":"ך","מ":"ם","נ":"ן","צ":"ץ"}
        if text[-1] in finals:
            text = text[:-1] + finals[text[-1]]
        return text

    def post_process(word, pos, lemma, morph):
        if word == lemma:
            if word + "\t" + pos in lex:
                if pos == "VERB" and "Fut" in morph:
                    lemma = lex[word + "\t" + pos]
                if pos == "VERB" and "Pres" in morph:
                    lemma = lex[word + "\t" + pos]
                if pos == "VERB" and "Part" in morph:
                    lemma = lex[word + "\t" + pos]
                if pos in ["NOUN", "ADJ"] and "Plur" in morph:
                    lemma = lex[word + "\t" + pos]
            else:
                if "Plur" in morph and pos in ["NOUN", "ADJ"] and (
                        word.endswith("ים") or word.endswith("ות")):
                    lemma = lemma[:-2]
                    if word.endswith("ות"):
                        lemma += "ה"
                    lemma = clean_final(lemma)
        return lemma

    uposed = [[l.split("\t") for l in s.split("\n")] for s in conllu.strip().split("\n\n")]
    dicts = CoNLL.convert_conll(uposed)
    for sent in dicts:
        for tok in sent:
            tok["id"] = int(tok["id"][0])
    doc = Document(dicts)
    lemmatized = lemmatizer(doc)
    output = []
    counter = 0
    for sent in lemmatized.sentences:
        for tok in sent.tokens:
            word = tok.words[0]
            lemma = word.lemma
            if lemmatizer.do_post_process:
                lemma = post_process(word.text, word.upos, word.lemma, morphs[counter])
            row = [str(word.id), word.text, lemma, word.upos, word.xpos, '_', str(word.head), "_", "_", "_"]
            output.append("\t".join(row))
            counter += 1
        output.append("")
    lemmatized = "\n".join(output)
    lemmatized = get_col(lemmatized,2)

    return lemmatized


def diaparse(parser, conllu):
    sents = conllu.strip().split("\n\n")
    parser_input = []

    for sent in sents:
        words = [(l.split("\t")[0],l.split("\t")[1]) for l in sent.split("\n") if "\t" in l]
        words = [w[1] for w in words if "." not in w[0] and "-" not in w[0]]
        parser_input.append(words)

    dataset = parser.predict(parser_input, prob=True)
    torch.cuda.empty_cache()

    out_parses = []
    for sent in dataset.sentences:
        out_parses.append(str(sent).strip())

    merged = inject_conllu("\n\n".join(out_parses) + "\n\n", conllu)

    return merged


def postprocess_morph(feats, words, lemmas, tags):
    def add_feat(morph, feat):
        if morph == "_":
            return feat
        else:
            attrs = morph.split("|")
            attrs = {a.split("=")[0]:a.split("=")[1] for a in attrs}
            attrs[feat.split("=")[0]] = feat.split("=")[1]
            return "|".join(sorted([a + "=" + attrs[a] for a in attrs]))

    output = []
    for i, lemma in enumerate(lemmas):
        word = words[i]
        feat = feats[i]
        tag = tags[i]
        if "HebBinyan" in feat or tag == "VERB":  # Rely on BERT to notice that binyan is needed, or if it's a VERB
            if (word,lemma) in binyan_lookup:
                feat = add_feat(feat,"HebBinyan=" + binyan_lookup[(word,lemma)])
            elif lemma in binyan_lemma_lookup:
                feat = add_feat(feat,"HebBinyan=" + binyan_lemma_lookup[lemma])
        output.append(feat)
    return output


def check_requirements():
    models_OK = True
    model_files = ["heb.sm" + str(sys.version_info[0]), "heb.xrm",
                   "heb.diaparser", "heb.sbdposmorph.pt",
                   "stanza" + os.sep + "he_lemmatizer.pt",
                   "stanza" + os.sep + "he_htb.pretrain.pt",
                   ]
    for model_file in model_files:
        if not os.path.exists(model_dir + model_file):
            sys.stderr.write("! Model file " + model_file + " missing in ./models/\n")
            models_OK = False

    return models_OK


def download_requirements(models_ok=True):
    urls = []
    if not models_ok:
        models_base = "http://gucorpling.org/amir/download/heb_models_v3/"
        urls.append(models_base + "heb.sm" + str(sys.version_info[0]))
        urls.append(models_base + "heb.diaparser")
        urls.append(models_base + "heb.xrm")
        urls.append(models_base + "he_htb.pretrain.pt")
        urls.append(models_base + "he_lemmatizer.pt")
        urls.append(models_base + 'heb.sbdposmorph.pt')
    for u in urls:
        sys.stderr.write("o Downloading from " + str(u) + "\n")
        base_name = u[u.rfind("/") + 1:]
        if "he_" in base_name:
            urlretrieve(u, model_dir + "stanza" + os.sep + base_name)
        else:
            urlretrieve(u, model_dir + base_name)
    sys.stderr.write("\n")


def nlp(input_data, do_whitespace=True, do_tok=True, do_tag=True, do_lemma=True, do_parse=True, do_entity=True,
        out_mode="conllu", sent_tag=None, preloaded=None,  from_pipes=False,cpu=False):

    data = input_data.replace("\t","")
    data = data.replace("\r","")

    if from_pipes:
        input_data = input_data.replace("|","")

    if preloaded is not None:
        rf_tok, xrenner, mtltagger,parser, lemmatizer = preloaded
    else:
        rf_tok = RFTokenizer(model=model_dir + "heb.sm" + str(sys.version_info[0]))
        xrenner = Xrenner(model=model_dir + "heb.xrm")
        parser = None if not do_parse else Parser.load(model_dir + "heb.diaparser",verbose=False)
        lemmatizer = None if not do_lemma and not do_tag else init_lemmatizer()
        mtltagger = Tagger(trainflag=False, bestmodelpath=model_dir, sequencelength=256,cpu=cpu)

    if do_whitespace:
        data = whitespace_tokenize(data, abbr=data_dir + "heb_abbr.tab",add_sents=sent_tag=="auto", from_pipes=from_pipes)

    if from_pipes:
        tokenized = data
    else:
        if do_tok:
            tokenized = rf_tok.rf_tokenize(data.strip().split("\n"))
            tokenized = "\n".join(tokenized)
        else:
            # Assume data is already one token per line
            tokenized = data

    bound_group_map = get_bound_group_map(tokenized) if out_mode == "conllu" else None

    if mtltagger:
        tagged_conllu, tokenized, morphs, words = mtltagger.predict(tokenized,sent_tag=sent_tag,checkpointfile=model_dir + 'heb.sbdposmorph.pt')

    if out_mode == "pipes":
        return tokenized
    else:
        tokenized = tokenized.split("\n")
        retokenized = []
        for line in tokenized:
            if line == '|':
                retokenized.append(line)
            else:
                retokenized.append("\n".join(line.split("|")))
        tokenized =  "\n".join(retokenized)

    if sent_tag == 'auto': sent_tag = 's'

    del mtltagger
    del rf_tok
    torch.cuda.empty_cache()

    if do_tag:
        zeros = ["0" for i in range(len(morphs))]
        zero_conllu = inject_col(zeros, tagged_conllu, into_col=6, skip_supertoks=True)
        lemmas = lemmatize(lemmatizer, zero_conllu, morphs)
        tagged = inject_col(tagged_conllu, tokenized, 4)

        if do_lemma:
            lemmatized = inject_col(lemmas, tagged, -1)
        else:
            lemmatized = tagged

        tags = [l.split("\t")[3] for l in tagged_conllu.split("\n") if "\t" in l]
        morphs = postprocess_morph(morphs, words, lemmas, tags)
        morphed = inject_col(morphs, lemmatized, -1)

        if not do_parse:
            if out_mode == "conllu":
                conllized = conllize(morphed, tag="PUNCT", element=sent_tag, no_zero=True, super_mapping=bound_group_map,
                                     attrs_as_comments=True)
                conllized = add_space_after(input_data,conllized)
                return conllized
            else:
                if not PY3:
                    morphed = morphed.decode("utf8")
                return morphed

    else:
        if out_mode == "conllu":
            conllized = conllize(tokenized, tag="PUNCT", element=sent_tag, no_zero=True, super_mapping=bound_group_map,
                                 attrs_as_comments=True)
            conllized = add_space_after(input_data,conllized)
            return conllized
        else:
            return tokenized

    if do_parse:
        conllized = conllize(morphed,tag="PUNCT",element=sent_tag,no_zero=True,super_mapping=bound_group_map,attrs_as_comments=True,ten_cols=True)
        parsed = diaparse(parser, conllized)
        parsed = morph_deped.run_depedit(parsed)

        if do_entity:
            xrenner.docname = "_"
            ents = xrenner.analyze(parsed,"conll_sent")
            ents = get_col(ents, -1)
            entified = inject_col(ents, parsed, col=-1, into_col=9, skip_supertoks=True)
            entified = add_space_after(input_data,entified)
            if PY3:
                return entified
            else:
                return entified.decode("utf8")
        else:
            parsed = add_space_after(input_data,parsed)
            return parsed
    else:
        if out_mode == "conllu":
            conllized = conllize(tagged,tag="PUNCT",element=sent_tag,no_zero=True,super_mapping=bound_group_map,attrs_as_comments=True)
            conllized = add_space_after(input_data,conllized)
            return conllized
        else:
            return tagged


def run_hebpipe():

    if sys.version_info[0] == 2 and sys.version_info[1] < 7:
        sys.stderr.write("Python versions below 2.7 are not supported.\n")
        sys.stderr.write("Your Python version:\n")
        sys.stderr.write(".".join([str(v) for v in sys.version_info[:3]]) + "\n")
        sys.exit(0)

    from argparse import ArgumentParser, RawDescriptionHelpFormatter

    parser = ArgumentParser(formatter_class=RawDescriptionHelpFormatter)
    parser.prog = "HebPipe - NLP Pipeline for Hebrew"
    parser.usage = "python heb_pipe.py [OPTIONS] files"
    parser.epilog = """Example usage:
--------------
Whitespace tokenize, tokenize morphemes, add pos, lemma, morph, dep parse with automatic sentence splitting, 
entity recognition and coref for one text file, output in default conllu format:
> python heb_pipe.py -wtpldec example_in.txt        

OR specify no processing options (automatically assumes you want all steps)
> python heb_pipe.py example_in.txt        

Just tokenize a file using pipes:
> python heb_pipe.py -wt -o pipes example_in.txt     

POS tag, lemmatize, add morphology and parse a pre-tokenized file, splitting sentences by existing <sent> tags:
> python heb_pipe.py -pld -s sent example_in.txt  

Add full analyses to a whole directory of *.txt files, output to a specified directory:    
> python heb_pipe.py -wtpldec --dirout /home/heb/out/ *.txt

Parse a tagged TT SGML file into CoNLL tabular format for treebanking, use existing tag <sent> to recognize sentence borders:
> python heb_pipe.py -d -s sent example_in.tt
"""
    parser.add_argument("files", help="File name or pattern of files to process (e.g. *.txt)")

    g1 = parser.add_argument_group("standard module options")
    g1.add_argument("-w","--whitespace", action="store_true", help='Perform white-space based tokenization of large word forms')
    g1.add_argument("-t","--tokenize", action="store_true", help='Tokenize large word forms into smaller morphological segments')
    g1.add_argument("-p","--posmorph", action="store_true", help='Do POS and Morph tagging')
    g1.add_argument("-l","--lemma", action="store_true", help='Do lemmatization')
    g1.add_argument("-d","--dependencies", action="store_true", help='Parse with dependency parser')
    g1.add_argument("-e","--entities", action="store_true", help='Add entity spans and types')
    g1.add_argument("-c","--coref", action="store_true", help='Add coreference annotations')
    g1.add_argument("-s","--sent", action="store", default="auto", help='XML tag to split sentences, e.g. sent for <sent ..> or none for no splitting (otherwise automatic sentence splitting)')
    g1.add_argument("-o","--out", action="store", choices=["pipes","conllu","sgml"], default="conllu", help='Output CoNLL format, SGML or just tokenize with pipes')

    g2 = parser.add_argument_group("less common options")
    g2.add_argument("-q","--quiet", action="store_true", help='Suppress verbose messages')
    g2.add_argument("-x","--extension", action="store", default='conllu', help='Extension for output files (default: .conllu)')
    g2.add_argument("--cpu", action="store_true", help='Use CPU instead of GPU (slower)')
    g2.add_argument("--disable_lex", action="store_true", help='Do not use lexicon during lemmatization')
    g2.add_argument("--dirout", action="store", default=".", help='Optional output directory (default: this dir)')
    g2.add_argument("--punct_sentencer", action="store_true", help='Only use punctuation (.?!) to split sentences (deprecated)')
    g2.add_argument("--from_pipes", action="store_true", help='Input contains subtoken segmentation with the pipe character (no automatic tokenization is performed)')
    g2.add_argument("--version", action="store_true", help='Print version number and quit')

    if "--version" in sys.argv:
        sys.stdout.write("HebPipe V" + __version__)
        sys.exit(1)

    opts = parser.parse_args()
    opts = diagnose_opts(opts)

    if opts.cpu:
        torch.cuda.is_available = lambda: False

    dotok = opts.tokenize

    if not opts.quiet:
        try:
            from .lib import timing
        except ImportError:  # direct script usage
            from lib import timing

    files = glob(opts.files)

    if not opts.quiet:
        log_tasks(opts)

    # Check if models are available
    if opts.posmorph or opts.lemma or opts.dependencies or opts.tokenize or opts.entities:
        models_OK = check_requirements()
        if not models_OK:
            sys.stderr.write("! You are missing required software:\n")
            if (opts.posmorph or opts.lemma):
                sys.stderr.write(" - Tagging, lemmatization and morphological analysis require models\n")
            if not models_OK:
                sys.stderr.write(" - Model files in models/ are missing\n")
            response = inp("Attempt to download missing files? [Y/N]\n")
            if response.upper().strip() == "Y":
                download_requirements(models_OK)
            else:
                sys.stderr.write("Aborting\n")
                sys.exit(0)

        lemmatizer = init_lemmatizer(cpu=opts.cpu, no_post_process=opts.disable_lex)
    else:
        lemmatizer = None

    if opts.posmorph:
        mtltagger = Tagger(trainflag=False, bestmodelpath=model_dir, sequencelength=256,cpu=opts.cpu)
    else:
        mtltagger = None

    if dotok:  # Pre-load stacked tokenizer for entire batch
        rf_tok = RFTokenizer(model=model_dir + "heb.sm" + str(sys.version_info[0]))
    else:
        rf_tok = None
    if opts.entities:  # Pre-load stacked tokenizer for entire batch
        xrenner = Xrenner(model=model_dir + "heb.xrm")
    else:
        xrenner = None

    dep_parser = Parser.load(model_dir+"heb.diaparser") if opts.dependencies else None

    for infile in files:
        base = os.path.basename(infile)
        if infile.endswith("." + opts.extension):
            outfile = base.replace("." + opts.extension,".out." + opts.extension)
        elif len(infile) > 4 and infile[-4] == ".":
            outfile = base[:-4] + "." + opts.extension
        else:
            outfile = base + "." + opts.extension

        if not opts.quiet:
            sys.stderr.write("Processing " + base + "\n")

        try:
            input_text = io.open(infile,encoding="utf8").read().replace("\r","")
        except UnicodeDecodeError:  # Fallback to support Windows Hebrew encoding
            input_text = io.open(infile,encoding="cp1255").read().replace("\r","")

        processed = nlp(input_text, do_whitespace=opts.whitespace, do_tok=dotok, do_tag=opts.posmorph, do_lemma=opts.lemma,
                               do_parse=opts.dependencies, do_entity=opts.entities, out_mode=opts.out,
                               sent_tag=opts.sent, preloaded=(rf_tok,xrenner,mtltagger,dep_parser,lemmatizer),
                                from_pipes=opts.from_pipes,cpu=opts.cpu)

        if len(files) > 1:
            with io.open(opts.dirout + os.sep + outfile, 'w', encoding="utf8", newline="\n") as f:
                if not PY3:
                    processed = unicode(processed)
                f.write((processed.strip() + "\n"))
        else:  # Single file, print to stdout
            if PY3:
                sys.stdout.buffer.write((processed+"\n\n").encode("utf8"))
            else:
                print(processed.encode("utf8"))

    fileword = " files\n\n" if len(files) > 1 else " file\n\n"
    sys.stderr.write("\nFinished processing " + str(len(files)) + fileword)


if __name__ == "__main__":
    import logging
    logging.disable(logging.INFO)
    run_hebpipe()




#!/usr/bin/python
# -*- coding: utf-8 -*-

from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.preprocessing import LabelEncoder
from collections import defaultdict
import numpy as np
import sys

PY3 = sys.version_info[0] == 3


def lambda_underscore():  # Module level named lambda-function to make defaultdict picklable
	return "_"


class LetterConfig:
	def __init__(self,letters=None, vowels=None, pos_lookup=None):

		if letters is None:
			self.letters = defaultdict(set)
			self.vowels = set()
			self.pos_lookup = defaultdict(lambda: "_")
		else:
			letter_cats = ["current_letter", "prev_prev_letter", "prev_letter", "next_letter", "next_next_letter", "prev_grp_first", "prev_grp_last", "next_grp_first", "next_grp_last"]
			self.letters = defaultdict(set)
			if "group_in_lex" in letters or "current_letter" in letters:  # Letters dictionary already instantiated - we are loading from disk
				self.letters.update(letters)
			else:
				for cat in letter_cats:
					self.letters[cat] = letters
			self.vowels = vowels
			self.pos_lookup = pos_lookup


class DataFrameSelector(BaseEstimator, TransformerMixin):
	def __init__(self, attribute_names):
		self.attribute_names = attribute_names
	def fit(self, X, y=None):
		return self
	def transform(self, X):
		return X[self.attribute_names].values


class MultiColumnLabelEncoder(LabelEncoder):
	"""
	Wraps sklearn LabelEncoder functionality for use on multiple columns of a
	pandas dataframe.

	"""

	def __init__(self, columns=None):
		self.encoder_dict = {}
		if isinstance(columns, list):
			self.columns = np.array(columns)
		else:
			self.columns = columns

	def fit(self, dframe):
		"""
		Fit label encoder to pandas columns.

		Access individual column classes via indexing `self.all_classes_`

		Access individual column encoders via indexing
		`self.all_encoders_`
		"""
		# if columns are provided, iterate through and get `classes_`
		if self.columns is not None:
			# ndarray to hold LabelEncoder().classes_ for each
			# column; should match the shape of specified `columns`
			self.all_classes_ = np.ndarray(shape=self.columns.shape, dtype=object)
			self.all_encoders_ = np.ndarray(shape=self.columns.shape, dtype=object)
			for idx, column in enumerate(self.columns):
				# fit LabelEncoder to get `classes_` for the column
				le = LabelEncoder()
				le.fit(dframe.loc[:, column].values)
				# append the `classes_` to our ndarray container
				self.all_classes_[idx] = (column, np.array(le.classes_.tolist(), dtype=object))
				# append this column's encoder
				self.all_encoders_[idx] = le
		else:
			# no columns specified; assume all are to be encoded
			self.columns = dframe.iloc[:, :].columns
			self.all_classes_ = np.ndarray(shape=self.columns.shape, dtype=object)
			for idx, column in enumerate(self.columns):
				le = LabelEncoder()
				le.fit(dframe.loc[:, column].values)
				self.all_classes_[idx] = (column, np.array(le.classes_.tolist(), dtype=object))
				self.all_encoders_[idx] = le
		return self

	def fit_transform(self, dframe):
		"""
		Fit label encoder and return encoded labels.

		Access individual column classes via indexing
		`self.all_classes_`

		Access individual column encoders via indexing
		`self.all_encoders_`

		Access individual column encoded labels via indexing
		`self.all_labels_`
		"""
		# if columns are provided, iterate through and get `classes_`
		if self.columns is not None:
			# ndarray to hold LabelEncoder().classes_ for each
			# column; should match the shape of specified `columns`
			self.all_classes_ = np.ndarray(shape=self.columns.shape, dtype=object)
			self.all_encoders_ = np.ndarray(shape=self.columns.shape, dtype=object)
			self.all_labels_ = np.ndarray(shape=self.columns.shape, dtype=object)
			for idx, column in enumerate(self.columns):
				# instantiate LabelEncoder
				le = LabelEncoder()
				# fit and transform labels in the column
				dframe.loc[:, column] = le.fit_transform(dframe.loc[:, column].values)
				self.encoder_dict[column] = le
				# append the `classes_` to our ndarray container
				self.all_classes_[idx] = (column, np.array(le.classes_.tolist(),dtype=object))
				self.all_encoders_[idx] = le
				self.all_labels_[idx] = le
		else:
			# no columns specified; assume all are to be encoded
			self.columns = dframe.iloc[:, :].columns
			self.all_classes_ = np.ndarray(shape=self.columns.shape,
										   dtype=object)
			for idx, column in enumerate(self.columns):
				le = LabelEncoder()
				dframe.loc[:, column] = le.fit_transform(
					dframe.loc[:, column].values)
				self.all_classes_[idx] = (column, np.array(le.classes_.tolist(), dtype=object))
				self.all_encoders_[idx] = le
		return dframe

	def transform(self, dframe):
		"""
		Transform labels to normalized encoding.
		"""
		if self.columns is not None:
			for idx, column in enumerate(self.columns):
				dframe.loc[:, column] = self.all_encoders_[idx].transform(dframe.loc[:, column].values)
		else:
			self.columns = dframe.iloc[:, :].columns
			for idx, column in enumerate(self.columns):
				dframe.loc[:, column] = self.all_encoders_[idx].transform(dframe.loc[:, column].values)
		return dframe

	def inverse_transform(self, dframe):
		"""
		Transform labels back to original encoding.
		"""
		if self.columns is not None:
			for idx, column in enumerate(self.columns):
				dframe.loc[:, column] = self.all_encoders_[idx] \
					.inverse_transform(dframe.loc[:, column].values)
		else:
			self.columns = dframe.iloc[:, :].columns
			for idx, column in enumerate(self.columns):
				dframe.loc[:, column] = self.all_encoders_[idx] \
					.inverse_transform(dframe.loc[:, column].values)
		return dframe


#@profile
def bg2array(bound_group, prev_group="", next_group="", print_headers=False, grp_id=-1, is_test=-1, config=None, train=False, freqs=None):

	output = []

	letters = config.letters
	vowels = config.vowels
	pos_lookup = config.pos_lookup

	group_in_lex = pos_lookup[bound_group] if pos_lookup[bound_group] in letters["group_in_lex"] or train else "_"

	for idx in range(len(bound_group)):

		char_feats = []
		so_far_substr = bound_group[:idx+1]
		remaining_substr = bound_group[idx+1:]
		remaining_substr_mns1 = bound_group[max(0,idx):]
		remaining_substr_mns2 = bound_group[max(0,idx-1):]
		so_far_pos = pos_lookup[so_far_substr] if pos_lookup[so_far_substr] in letters["so_far_pos"] or train else "_"
		remaining_pos = pos_lookup[remaining_substr] if pos_lookup[remaining_substr] in letters["remaining_pos"] or train else "_"
		remaining_pos_mns1 = pos_lookup[remaining_substr_mns1] if pos_lookup[remaining_substr_mns1] in letters["remaining_pos_mns1"] or train else "_"
		remaining_pos_mns2 = pos_lookup[remaining_substr_mns2] if pos_lookup[remaining_substr_mns2] in letters["remaining_pos_mns2"] or train else "_"

		current_letter = bound_group[idx] if bound_group[idx] in letters["current_letter"] or train else "_"
		current_vowel = 1 if bound_group[idx] in vowels else 0
		if idx > 0:
			prev_letter = bound_group[idx-1] if bound_group[idx-1] in letters["prev_letter"] or train else "_"
			prev_vowel = 1 if bound_group[idx-1] in vowels else 0
		else:
			prev_letter = "_"
			prev_vowel = -1
		if idx > 1:
			prev_prev_letter = bound_group[idx-2] if bound_group[idx-2] in letters["prev_prev_letter"] else"_"
			prev_prev_vowel = 1 if bound_group[idx-2] in vowels else 0
		else:
			prev_prev_letter = "_"
			prev_prev_vowel = -1
		if idx < len(bound_group)-1:
			next_letter = bound_group[idx+1] if bound_group[idx+1] in letters["next_letter"] else"_"
			next_vowel = 1 if bound_group[idx+1] in vowels else 0
		else:
			next_letter = "_"
			next_vowel = -1
		if idx < len(bound_group)-2:
			next_next_letter = bound_group[idx+2] if bound_group[idx+2] in letters["next_next_letter"] or train else "_"
			next_next_vowel = 1 if bound_group[idx+2] in vowels else 0
		else:
			next_next_letter = "_"
			next_next_vowel = -1

		char_feats += [idx, group_in_lex, current_letter, prev_prev_letter, prev_letter, next_letter, next_next_letter, len(bound_group)]
		char_feats += [current_vowel, prev_prev_vowel, prev_vowel, next_vowel, next_next_vowel, so_far_pos, remaining_pos, remaining_pos_mns1, remaining_pos_mns2]
		headers = ["idx","group_in_lex","current_letter","prev_prev_letter","prev_letter","next_letter","next_next_letter","len_bound_group"]
		headers += ["current_vowel","prev_prev_vowel","prev_vowel","next_vowel","next_next_vowel", "so_far_pos", "remaining_pos","remaining_pos_mns1","remaining_pos_mns2"]

		# POS lookup features
		all_pos_feats = []
		for chargram_idx, prev_char in enumerate([-4,-3,-2,-1,1,2,3,4]):
			substr = ""
			header_prefix = "_"
			if prev_char < 0:
				header_prefix = "mns" + str(abs(prev_char)) + "_"
				if idx + prev_char >= 0:
					substr = bound_group[idx+prev_char:idx+1]
			elif prev_char > 0:
				header_prefix = "pls" + str(abs(prev_char)) + "_"
				if idx + prev_char <= len(bound_group):
					substr = bound_group[idx:idx+prev_char+1]

			coarse_tag = pos_lookup[substr] if pos_lookup[substr] in letters[header_prefix + "coarse"] or train else "_"

			pos_feats = [coarse_tag]
			if print_headers:
				headers.append(header_prefix + "coarse")
			all_pos_feats += pos_feats #clean_pos_feats

		char_feats += all_pos_feats

		if prev_group != "":
			prev_first = prev_group[0] if prev_group[0] in letters["prev_grp_first"] or train else "_"
			prev_last = prev_group[-1] if prev_group[-1] in letters["prev_grp_last"] or train else "_"
			prev_grp_pos = pos_lookup[prev_group] if pos_lookup[prev_group] in letters["prev_grp_pos"] or train else "_"
			prev_feats = [prev_first,prev_last,len(prev_group),prev_grp_pos]
			char_feats += prev_feats
			if print_headers:
				headers += ["prev_grp_first","prev_grp_last","prev_grp_len","prev_grp_pos"]
		if next_group != "":
			next_first = next_group[0] if next_group[0] in letters["next_grp_first"] or train else "_"
			next_last = next_group[-1] if next_group[-1] in letters["next_grp_last"] or train else "_"
			next_grp_pos = pos_lookup[next_group] if pos_lookup[next_group] in letters["next_grp_pos"] or train else "_"
			next_feats = [next_first,next_last,len(next_group),next_grp_pos]
			char_feats += next_feats
			if print_headers:
				headers += ["next_grp_first","next_grp_last","next_grp_len","next_grp_pos"]

		headers += ["freq_ratio"]
		if freqs is None:
			char_feats += [0.0]
		else:
			f_sofar = freqs[so_far_substr]
			f_remain = freqs[remaining_substr]
			f_whole = freqs[bound_group] + 0.0000000001  # Delta smooth whole
			char_feats += [f_sofar*f_remain/f_whole]

		if grp_id > -1:
			char_feats += [grp_id]
			if print_headers:
				headers += ["grp_id"]
		if is_test > -1:
			char_feats += [is_test]
			if print_headers:
				headers += ["is_test"]

		char_feats += [bound_group,prev_group,next_group]
		headers += ["this_group","prev_group","next_group"]

		output.append(char_feats)


	if print_headers:
		return headers
	else:
		return output


def segs2array(segs):
	output = []
	cursor = 0
	while cursor < len(segs) - 1:  # Word not over yet
		if segs[cursor + 1] == "|":
			output.append(1)  # 1 = positive class
			cursor += 2
		else:
			output.append(0)  # 0 = negative class
			cursor += 1
	output.append(0)

	return output


